# coding=utf-8
# -*- Python Version: 2.7 -*-

"""Module for reading in HBJSON and converting to PHX-Model.

Running the 'convert_hbjson_to_PHX' function will call a new subprocess using the 
Ladybug Tools Python 3.7 interpreter.
"""

from __future__ import division

import os
import subprocess

try:
    from typing import Any, Dict, List, Tuple, Union
except ImportError:
    pass  # Python 2.7

try:  # import the core honeybee dependencies
    from honeybee.config import folders as hb_folders
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))


def _run_subprocess(commands):
    # type: (List[str]) -> Tuple[Any, Any]
    """Run a python subprocess.Popen, using the supplied commands

    Arguments:
    ----------
        * commands: (List[str]): A list of the commands to pass to Popen

    Returns:
    --------
        * Tuple:
            - [0]: stdout
            - [1]: stderr
    """
    # -- Create a new PYTHONHOME to avoid the Rhino-8 issues
    CUSTOM_ENV = os.environ.copy()
    CUSTOM_ENV["PYTHONHOME"] = ""

    use_shell = True if os.name == "nt" else False

    process = subprocess.Popen(
        commands,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=use_shell,
        env=CUSTOM_ENV,
    )

    stdout, stderr = process.communicate()

    if stderr:
        if "Defaulting to Windows directory." in str(stderr):
            print("WARNING: {}".format(stderr))
        else:
            print(stderr)
            raise Exception(stderr)

    for _ in str(stdout).split("\\n"):
        print(_)

    return stdout, stderr


def _run_subprocess_from_shell(commands):
    # type: (List[str]) -> Tuple[Any, Any]
    """Run a python subprocess.Popen THROUGH a MacOS terminal via a shell, using the supplied commands.

    When talking to Excel on MacOS it is necessary to run through a Terminal since Rhino
    cannot get the right 'permissions' to interact with Excel. This is a workaround and not
    required on Windows OS.

    Arguments:
    ----------
        * _commands: (List[str]): A list of the commands to pass to Popen

    Returns:
    --------
        * Tuple:
            - [0]: stdout
            - [1]: stderr
    """

    # -- Create a new PYTHONHOME to avoid the Rhino-8 issues
    CUSTOM_ENV = os.environ.copy()
    CUSTOM_ENV["PYTHONHOME"] = ""

    use_shell = True if os.name == "nt" else False

    # -- Make sure the files are executable
    shell_file = commands[0]
    try:
        subprocess.check_call(["chmod", "u+x", shell_file])
    except Exception as e:
        print("Failed to make the shell file executable: {}".format(e))
        raise e

    python_script_path = commands[3]
    try:
        subprocess.check_call(["chmod", "u+x", python_script_path])
    except Exception as e:
        print("Failed to make the python script executable: {}".format(e))
        raise e

    process = subprocess.Popen(
        commands,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=use_shell,
        env=CUSTOM_ENV,
    )

    stdout, stderr = process.communicate()

    if stderr:
        if "Defaulting to Windows directory." in str(stderr):
            print("Warning: {}".format(stderr))
        else:
            raise Exception(stderr)

    for _ in str(stdout).split("\\n"):
        print(_)

    return stdout, stderr


def convert_hbjson_to_WUFI_XML(
    _hbjson_file,
    _save_file_name,
    _save_folder,
    _group_components=True,
    _merge_faces=False,
    _merge_spaces_by_erv=False,
    _log_level=0,
    *args,
    **kwargs
):
    # type: (str, str, str, bool, Union[bool, float], bool, int, List, Dict) -> tuple[str, str, str, str]
    """Read in an hbjson file and output a new WUFI XML file in the designated location.

    Arguments:
    ---------
        * _hbjson (str): File path to an HBJSON file to be read in and converted to a PHX-Model.
        * _save_file_name (str): The XML filename.
        * _save_folder (str): The folder to save the new XML file in.
        * _group_components (bool): Group components by construction? Default=True
        * _merge_faces (bool | float): Merge together faces of the same type and touching? If
            a number is provided, it will be used as the tolerance when merging faces. Default=False
        * _merge_spaces_by_erv (bool): Merge spaces that are connected by ERVs? Default=False
        * _log_level (int): Set the logging level for the subprocess. Default=0
        * args (List): Additional arguments to pass to the subprocess.
        * kwargs (Dict): Additional keyword arguments to pass to the subprocess.

    Returns:
    --------
        * tuple
            - [0] (str): The path to the output XML file.
            - [1] (str): The output xml filename.
            - [2] (str): The stdout from the subprocess.
            - [3] (str): The stderr from the subprocess.
    """

    # -- Specify the path to the subprocess python script to run
    run_file_path = os.path.join(hb_folders.python_package_path, "PHX", "hbjson_to_wufi_xml.py")

    # -- check the file paths
    assert os.path.isfile(_hbjson_file), "No HBJSON file found at {}.".format(_hbjson_file)
    assert os.path.isfile(run_file_path), "No Python file to run found at: {}".format(run_file_path)

    # -------------------------------------------------------------------------
    # -- Read in the HBJSON, convert to WUFI XML
    print("Using python interpreter: '{}'".format(hb_folders.python_exe_path))
    print("Running py script: '{}' Using HBJSON file: '{}'".format(run_file_path, _hbjson_file))
    commands = [
        hb_folders.python_exe_path,
        run_file_path,
        _hbjson_file,
        _save_file_name,
        _save_folder,
        str(_group_components),
        str(_merge_faces),
        str(_merge_spaces_by_erv),
        str(_log_level),
    ]
    stdout, stderr = _run_subprocess(commands)

    # -------------------------------------------------------------------------
    # -- return the dir and filename of the xml created
    return _save_folder, _save_file_name, stdout, stderr


def write_hbjson_to_phpp(_hbjson_file, _lbt_python_site_packages_path, _activate_variants="False"):
    # type: (str, str, str) -> Tuple[Any, Any]
    """Read in an hbjson file and write out to a PHPP file.

    Arguments:
    ---------
        * _hbjson (str): File path to an HBJSON file to be read in and converted to a PHX-Model.

        * _lbt_python_site_packages_path (str): The full path to the LBT Python-3 Site-packages
            folder where PHX and Honeybee-PH are installed

        * _activate_variants (str): Default="False". Set True if you would like to
            connect all of the various PHPP inputs to the 'Variants' worksheet. This
            is used when testing various combinations of attributes during the
            early design phase. Note that if activated, any inputs will get overwritten
            when the connection to the 'Variants' worksheet is made.
            Note: Args must be strings, not actual boolean True/False.
    Returns:
    --------
        * Tuple:
            - [0]: stdout
            - [1]: stderr
    """

    # -- Specify the path to the subprocess python script to run
    run_file_path = os.path.join(hb_folders.python_package_path, "PHX", "hbjson_to_phpp.py")

    # -- check the file paths
    if not os.path.isfile(_hbjson_file):
        raise Exception("\nNo HBJSON file found at {}?".format(_hbjson_file))
    if not os.path.isfile(run_file_path):
        raise Exception("\nNo Python file to run found at: {}?".format(run_file_path))
    if not os.path.exists(_lbt_python_site_packages_path):
        raise Exception("\nNo Python site_packages folder found at: {}?".format(_lbt_python_site_packages_path))

    # -------------------------------------------------------------------------
    # -- Read in the HBJSON, write out to PHPP
    if os.name == "nt":
        commands = [
            hb_folders.python_exe_path,
            run_file_path,
            _hbjson_file,
            _lbt_python_site_packages_path,
            _activate_variants,
        ]
        stdout, stderr = _run_subprocess(commands)
    else:
        # -- If on MacOS, run the subprocess through a shell
        # -- and another terminal window in order to connect to Excel.
        # -- This is a workaround for the permissions issue on MacOS.
        # -- See:
        # -- https://discourse.mcneel.com/t/python-subprocess-permissions-error-on-mac-os-1743/142830/6
        # -- Find the file paths to run
        shell_file = os.path.join(_lbt_python_site_packages_path, "PHX", "_hbjson_to_phpp.sh")
        execution_root = os.path.join(_lbt_python_site_packages_path, "PHX")
        python_path = hb_folders.python_exe_path
        python_script_path = os.path.join(_lbt_python_site_packages_path, "PHX", "hbjson_to_phpp.py")
        hbjson_file = _hbjson_file
        # -- Build up the commands to run
        commands = [
            shell_file,
            execution_root,
            python_path,
            python_script_path,
            hbjson_file,
            _activate_variants,
        ]
        stdout, stderr = _run_subprocess_from_shell(commands)

    return stdout, stderr



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Building Classes"""

from __future__ import annotations

import operator
from collections import defaultdict
from dataclasses import dataclass, field
from functools import reduce
from typing import Any, ClassVar, Dict, Generator, List, NamedTuple, Optional, Sequence, Set, Union, ValuesView

from PHX.model import elec_equip, geometry, spaces
from PHX.model.components import PhxApertureElement, PhxComponentAperture, PhxComponentOpaque, PhxComponentThermalBridge
from PHX.model.enums.building import (
    AttachedZoneType,
    SpecificHeatCapacity,
    ZoneType,
    WufiVolumeGrossMode,
    WufiVolumeNetMode,
    WufiWeightedFloorAreaMode,
)
from PHX.model.hvac import collection
from PHX.model.programs import occupancy


@dataclass
class PhxZone:
    _count: ClassVar[int] = 0
    display_name: str = ""
    volume_gross_mode: WufiVolumeGrossMode = WufiVolumeGrossMode.FROM_VISUALIZED_COMPONENTS
    volume_gross: float = 0.0
    volume_net_mode: WufiVolumeNetMode = WufiVolumeNetMode.USER_DEFINED
    volume_net: float = 0.0
    weighted_floor_area_mode: WufiWeightedFloorAreaMode = WufiWeightedFloorAreaMode.USER_DEFINED
    weighted_net_floor_area: float = 0.0
    clearance_height: float = 2.5
    res_occupant_quantity: float = 0.0
    res_number_bedrooms: int = 0
    res_number_dwellings: int = 0
    specific_heat_capacity: SpecificHeatCapacity = SpecificHeatCapacity.LIGHTWEIGHT

    # -- Neeed for WUFI's Attached Zones
    zone_type: ZoneType = ZoneType.SIMULATED
    attached_zone_type = AttachedZoneType.NONE
    attached_zone_reduction_factor: float = 1.0
    id_num: int = field(init=False, default=0)

    spaces: List[spaces.PhxSpace] = field(default_factory=list)
    _thermal_bridges: Dict[str, PhxComponentThermalBridge] = field(default_factory=dict)

    # TODO: see if this can be safely removed?
    # ---------------------------------------------------------
    occupancy: Optional[occupancy.PhxProgramOccupancy] = None
    # ---------------------------------------------------------

    lighting: Any = None
    elec_equipment_collection: elec_equip.PhxElectricDeviceCollection = field(
        default_factory=elec_equip.PhxElectricDeviceCollection
    )
    exhaust_ventilator_collection: collection.PhxExhaustVentilatorCollection = field(
        default_factory=collection.PhxExhaustVentilatorCollection
    )

    # ---------------------------------------------------------
    # -- Allow user-overrides of the weighted floor area
    tfa_override: Optional[float] = None
    icfa_override: Optional[float] = None

    # -- This flag gets stored here since we don't really have any other way to
    # -- pass these kinds of 'settings' down to the XML writer.
    merge_spaces_by_erv: bool = False

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def add_thermal_bridge(self, _thermal_bridge: PhxComponentThermalBridge) -> None:
        """Add a new PhxComponentThermalBridge to the PhxZone."""
        self._thermal_bridges[str(_thermal_bridge.identifier)] = _thermal_bridge

    def add_thermal_bridges(
        self,
        _thermal_bridges: Union[PhxComponentThermalBridge, Sequence[PhxComponentThermalBridge]],
    ) -> None:
        """Add a new PhxComponentThermalBridge (or list of Bridges) to the PhxZone."""
        if not isinstance(_thermal_bridges, Sequence):
            _thermal_bridges = (_thermal_bridges,)

        for tb in _thermal_bridges:
            self._thermal_bridges[str(tb.identifier)] = tb

        return None

    def clear_thermal_bridges(self) -> None:
        """Clear all of the PhxComponentThermalBridge objects from the PhxZone."""
        self._thermal_bridges = dict()

    @property
    def thermal_bridges(self) -> ValuesView[PhxComponentThermalBridge]:
        """Return all of the PhxComponentThermalBridge objects in the PhxZone."""
        return self._thermal_bridges.values()

    @property
    def ventilated_spaces(self) -> List[spaces.PhxSpace]:
        """Return a list of all the spaces in the PhxZone which have some amount of ventilation airflow."""
        return [s for s in self.spaces if s.has_ventilation_airflow]

    @property
    def ventilated_spaces_grouped_by_erv(self) -> list[list[spaces.PhxSpace]]:
        """Return a dictionary of spaces grouped by their ERV ID."""
        # -- Get all the spaces, sorted by ERV-id-number
        grouped_spaces = defaultdict(list)
        for s in self.ventilated_spaces:
            grouped_spaces[s.vent_unit_id_num].append(s)

        # -- Return the spaces as a list of lists, sorted by the ERV-id-number
        spaces_ = []
        for k in sorted(grouped_spaces.keys()):
            spaces_.append(grouped_spaces[k])
        return spaces_

    def merge_thermal_bridges(self) -> None:
        """Merge together all the Thermal Bridges in the Zone if they have the same 'unique_key' attribute."""
        # -- Group the thermal bridges by their unique key
        unique_tbs: defaultdict[str, List[PhxComponentThermalBridge]] = defaultdict(list)
        for tb in self.thermal_bridges:
            unique_tbs[tb.unique_key].append(tb)

        # -- Create a new merged TB from each of the groups
        merged_tb_components: List[PhxComponentThermalBridge] = []
        for tb_group in unique_tbs.values():
            merged_tb_components.append(reduce(operator.add, tb_group))

        # -- Reset the Zone's Thermal Bridges
        self.clear_thermal_bridges()
        for tb in merged_tb_components:
            self.add_thermal_bridge(tb)


@dataclass
class PhxBuilding:
    """PHX Building Class"""

    # -- Only opaque components (and shades) are stored in the _components list
    # -- as the apertures are stored in the opaque components themselves
    _components: List[PhxComponentOpaque] = field(default_factory=list)
    zones: List[PhxZone] = field(default_factory=list)

    @property
    def weighted_net_floor_area(self) -> float:
        """Returns the total weighted net floor area of all zones in the PhxBuilding."""
        return sum(z.weighted_net_floor_area for z in self.zones)

    @property
    def tfa_override(self) -> Optional[float]:
        """Return the total floor area override of the PhxBuilding."""
        if all(z.tfa_override is None for z in self.zones):
            return None
        return sum(z.tfa_override for z in self.zones if z.tfa_override is not None)

    @property
    def net_volume(self) -> float:
        """Returns the total net-volume of all the zones in the PhxBuilding."""
        return sum(z.volume_net for z in self.zones)

    def add_components(self, _components: Union[PhxComponentOpaque, Sequence[PhxComponentOpaque]]) -> None:
        """Add new PHX Components to the PhxBuilding."""
        if not isinstance(_components, Sequence):
            _components = (_components,)

        for compo in _components:
            self._components.append(compo)

    def add_component(self, _component: PhxComponentOpaque) -> None:
        """Add a new PHX Components to the PhxBuilding."""
        self._components.append(_component)

    def add_zones(self, _zones: Union[PhxZone, Sequence[PhxZone]]) -> None:
        """Add a new PhxZone to the PhxBuilding."""
        if not isinstance(_zones, Sequence):
            _zones = (_zones,)

        for zone in _zones:
            self.zones.append(zone)

    def add_zone(self, _zone: PhxZone) -> None:
        """Add a new PhxZone to the PhxBuilding."""
        self.zones.append(_zone)

    def merge_opaque_components_by_assembly(self) -> None:
        """Merge together all the Opaque-Components in the Building if they gave the same Attributes."""
        # -- Group the opaque components by their unique key / type
        new_component_groups = defaultdict(list)
        for c in self.opaque_components:
            new_component_groups[c.unique_key].append(c)

        # -- Create new components from the group
        grouped_opaque_components: List[PhxComponentOpaque] = []
        for component_group in new_component_groups.values():
            grouped_opaque_components.append(reduce(operator.add, component_group))

        # -- Re-Set the names of all the new components
        # -- This is already done for components that __add__, but single surfaces that skipped
        # -- the __add__ method also need to have their names reset.
        for component in grouped_opaque_components:
            component.display_name = f"{component.face_type.name} [{component.assembly.display_name}]"

        # -- Reset the Building's Components
        self._components = grouped_opaque_components

    def merge_aperture_components_by_assembly(self) -> None:
        """Merge together all the Aperture-Components in the Building if they have the same Attributes."""
        # -- Group the aperture components by their unique key / type
        new_components: List[PhxComponentOpaque] = []
        for c in self.opaque_components:
            new_component_groups = defaultdict(list)
            for a in c.apertures:
                new_component_groups[a.unique_key].append(a)

            # -- Create new components from the groups
            grouped_aperture_components: List[PhxComponentAperture] = []
            for component_group in new_component_groups.values():
                grouped_aperture_components.append(reduce(operator.add, component_group))

            # -- Reset the Components's Apertures
            c.apertures = grouped_aperture_components
            new_components.append(c)

        # -- Reset the Building's Components
        self._components = new_components

    def merge_thermal_bridges(self) -> None:
        """Merge together all the Thermal Bridges in each of the Building's Zones if they have the same Attributes."""
        for z in self.zones:
            z.merge_thermal_bridges()

    @property
    def all_components(self) -> List[Union[PhxComponentOpaque, PhxComponentAperture]]:
        """Return a list of all the Opaque and Aperture Components in the Building.

        Returns:
        --------
            * (List[Union[PhxComponentOpaque, PhxComponentAperture]]) A list of all
                the opaque and aperture components.
        """
        all_components: List[Union[PhxComponentOpaque, PhxComponentAperture]] = []
        all_components = [c for c in self._components]
        all_components += self.aperture_components
        return sorted(all_components, key=lambda c: c.id_num)

    @property
    def aperture_components(self) -> List[PhxComponentAperture]:
        """Returns a sorted list (by display name) of all the aperture components in the building.

        Returns:
        --------
            * (List[PhxComponentAperture]) A sorted list of all the aperture components.
        """

        # -- An aperture might be 'in' multiple components, so collect each unique one
        unique_apertures: Dict[int, PhxComponentAperture] = {}
        for c in self.opaque_components:
            for ap in c.apertures:
                unique_apertures[id(ap)] = ap

        return sorted((ap for ap in unique_apertures.values()), key=lambda ap: ap.display_name)

    @property
    def aperture_elements(self) -> List[PhxApertureElement]:
        """Returns a sorted list (by display name) of all the aperture elements in the building.

        Returns:
        --------
            * (List[PhxApertureElement]) A sorted list of all the aperture components.
        """
        return sorted(
            [el for ap in self.aperture_components for el in ap.elements],
            key=lambda el: el.display_name,
        )

    @property
    def aperture_elements_by_orientation(self):
        """Return all of the Aperture Elements, grouped by their cardinal orientation."""

        class Output(NamedTuple):
            north: List[PhxApertureElement]
            east: List[PhxApertureElement]
            south: List[PhxApertureElement]
            west: List[PhxApertureElement]
            horizontal: List[PhxApertureElement]

        aperture_elements_by_orientation = Output([], [], [], [], [])

        for el in self.aperture_elements:
            if not el.polygon:
                continue

            if el.polygon.angle_from_horizontal <= 45 or el.polygon.angle_from_horizontal >= 135:
                aperture_elements_by_orientation.horizontal.append(el)
            elif el.polygon.cardinal_orientation_angle < 45:
                aperture_elements_by_orientation.north.append(el)
            elif el.polygon.cardinal_orientation_angle < 135:
                aperture_elements_by_orientation.east.append(el)
            elif el.polygon.cardinal_orientation_angle < 225:
                aperture_elements_by_orientation.south.append(el)
            elif el.polygon.cardinal_orientation_angle < 315:
                aperture_elements_by_orientation.west.append(el)
            else:
                aperture_elements_by_orientation.north.append(el)

        return aperture_elements_by_orientation

    @property
    def aperture_components_horizontal(self) -> List[PhxComponentAperture]:
        return [ap for ap in self.aperture_components]

    @property
    def wall_aperture_components(self) -> List[PhxComponentAperture]:
        """Returns a sorted list (by display name) of all the wall aperture (window) components in the building.

        Returns:
        --------
            * (List[PhxComponentAperture]) A sorted list of all the wall aperture components.
        """
        # -- An aperture might be 'in' multiple components, so collect each unique one
        unique_apertures: Dict[int, PhxComponentAperture] = {}
        for c in self.above_grade_wall_components:
            for ap in c.apertures:
                unique_apertures[id(ap)] = ap

        return sorted((ap for ap in unique_apertures.values()), key=lambda ap: ap.display_name)

    @property
    def roof_aperture_components(self) -> List[PhxComponentAperture]:
        """Returns a sorted list (by display name) of all the roof aperture (skylight) components in the building.

        Returns:
        --------
            * (List[PhxComponentAperture]) A sorted list of all the roof aperture components.
        """
        # -- An aperture might be 'in' multiple components, so collect each unique one
        unique_apertures: Dict[int, PhxComponentAperture] = {}
        for c in self.roof_components:
            for ap in c.apertures:
                unique_apertures[id(ap)] = ap

        return sorted((ap for ap in unique_apertures.values()), key=lambda ap: ap.display_name)

    @property
    def opaque_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the opaque non-shade components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the opaque components.
        """
        return sorted([c for c in self._components if not c.is_shade], key=lambda _: _.display_name)

    @property
    def roof_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the roof components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the roof components.
        """
        return sorted([c for c in self._components if c.is_roof], key=lambda _: _.display_name)

    @property
    def above_grade_wall_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the above grade wall components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the above grade wall components.
        """
        return sorted(
            [c for c in self._components if c.is_above_grade_wall],
            key=lambda _: _.display_name,
        )

    @property
    def shading_components(self) -> List[PhxComponentOpaque]:
        """Returns a list of all the opaque shade components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the shading components.
        """
        return sorted([c for c in self._components if c.is_shade], key=lambda _: _.display_name)

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon IDs of all Polygons from all the Components in the building."""
        p_ids = set()
        for compo in self.all_components:
            p_ids.update(compo.polygon_ids)
        return p_ids

    @property
    def polygons(self) -> List[geometry.PhxPolygon]:
        """Returns a list of all the Polygons of all the Components in the building."""
        return [poly for component in self.all_components for poly in component.polygons]

    @property
    def all_spaces(self) -> List[spaces.PhxSpace]:
        """Return a list of all the Spaces in the Building."""
        return [s for z in self.zones for s in z.spaces]

    def __bool__(self) -> bool:
        return bool(self.opaque_components) or bool(self.zones)

    def get_total_gross_wall_area(self) -> float:
        """Returns the total wall area of all the opaque components in the building."""
        return sum(c.get_total_gross_component_area() for c in self.above_grade_wall_components)

    def get_total_net_wall_area(self) -> float:
        """Returns the total net wall area of all the opaque components in the building."""
        return sum(c.get_total_net_component_area() for c in self.above_grade_wall_components)

    def get_total_gross_roof_area(self) -> float:
        """Returns the total roof area of all the opaque components in the building."""
        return sum(c.get_total_gross_component_area() for c in self.roof_components)

    def get_total_net_roof_area(self) -> float:
        """Returns the total net roof area of all the opaque components in the building."""
        return sum(c.get_total_net_component_area() for c in self.roof_components)

    def get_total_wall_aperture_area(self) -> float:
        """Returns the total window area of all the opaque components in the building."""
        return sum(c.get_total_aperture_area() for c in self.wall_aperture_components)

    def get_total_roof_aperture_area(self) -> float:
        """Returns the total skylight area of all the opaque components in the building."""
        return sum(c.get_total_aperture_area() for c in self.roof_aperture_components)

    def scale_all_wall_aperture_components(self, _scale_factor: float = 1.0) -> None:
        """Scale all the wall-aperture component's polygons by a given factor.

        Args:
        -----
            * scale_factor (float): The factor to scale the aperture components by.
        """
        for c in self.wall_aperture_components:
            c.scale(_scale_factor)

    def scale_all_roof_aperture_components(self, _scale_factor: float = 1.0) -> None:
        """Scale all the roof-aperture component's polygons by a given factor.

        Args:
        -----
            * scale_factor (float): The factor to scale the aperture components by.
        """
        for c in self.roof_aperture_components:
            c.scale(_scale_factor)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Project Classes"""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Set, Tuple

from PHX.model.building import PhxBuilding, PhxZone
from PHX.model.certification import PhxPhiCertification, PhxPhiusCertification
from PHX.model.constructions import PhxConstructionOpaque, PhxConstructionWindow
from PHX.model.geometry import PhxGraphics3D
from PHX.model.hvac import PhxMechanicalDevice
from PHX.model.hvac.collection import NoDeviceFoundError, PhxMechanicalSystemCollection
from PHX.model.phx_site import PhxSite
from PHX.model.schedules import lighting, occupancy, ventilation
from PHX.model.shades import PhxWindowShade
from PHX.model.utilization_patterns import (
    UtilizationPatternCollection_Lighting,
    UtilizationPatternCollection_Occupancy,
    UtilizationPatternCollection_Ventilation,
)


@dataclass
class WufiPlugin:
    insert_plugin: bool = False
    name_dll: Optional[Any] = None
    status_plugin: Optional[Any] = None


@dataclass
class PhxVariant:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    name: Optional[str] = "unnamed_variant"
    remarks: Optional[str] = None
    plugin: Optional[WufiPlugin] = field(default_factory=WufiPlugin)
    building: PhxBuilding = field(default_factory=PhxBuilding)
    phius_cert: PhxPhiusCertification = field(default_factory=PhxPhiusCertification)
    phi_cert: PhxPhiCertification = field(default_factory=PhxPhiCertification)
    site: PhxSite = field(default_factory=PhxSite)

    # -- Allow for multiple mechanical 'collections' in a variant
    # -- If WUFI, these are called 'systems', but they also use
    # -- the word 'system' in other places. So to avoid confusion, lets
    # -- call them 'collections' here
    _mech_collections: List[PhxMechanicalSystemCollection] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

        # -- Always Add a default mech-collection
        self._mech_collections.append(PhxMechanicalSystemCollection())

    @property
    def mech_systems(self) -> PhxMechanicalSystemCollection:
        """Return the Default Mechanical System Collection for the variant.

        This is a facade to allow for backwards compatibility as well.
        """
        print(
            "Warning: You should be using the 'mech_collections' property "
            "to access PHX mechanical devices instead of 'mech_systems'."
        )
        return self._mech_collections[0]

    @property
    def mech_collections(self) -> List[PhxMechanicalSystemCollection]:
        """Return the list of Mechanical System Collections for the variant."""
        return self._mech_collections

    @property
    def default_mech_collection(self) -> PhxMechanicalSystemCollection:
        """Return the Default Mechanical System Collection for the variant."""
        return self._mech_collections[0]

    @property
    def graphics3D(self):
        """Collects all of the geometry (Polygons, Vertices) in the Project."""
        phx_graphics3D = PhxGraphics3D()
        for phx_component in self.building.all_components:
            phx_graphics3D.add_polygons(phx_component.polygons)
        return phx_graphics3D

    @property
    def phi_certification_major_version(self) -> int:
        """Return the PHI Certification Version Number."""
        return self.phi_cert.version

    @property
    def zones(self) -> List[PhxZone]:
        """Return a list of all the PHX Zones in the variant.building"""
        return self.building.zones

    def get_total_gross_wall_area(self) -> float:
        """Returns the total wall area of the variant.building"""
        return self.building.get_total_gross_wall_area()

    def get_total_net_wall_area(self) -> float:
        """Returns the total net wall area of the variant.building"""
        return self.building.get_total_net_wall_area()

    def get_total_wall_aperture_area(self) -> float:
        """Returns the total window area of the variant.building"""
        return self.building.get_total_wall_aperture_area()

    def get_total_gross_roof_area(self) -> float:
        """Returns the total wall area of the variant.building"""
        return self.building.get_total_gross_roof_area()

    def get_total_net_roof_area(self) -> float:
        """Returns the total net wall area of the variant.building"""
        return self.building.get_total_net_roof_area()

    def get_total_roof_aperture_area(self) -> float:
        """Returns the total window area of the variant.building"""
        return self.building.get_total_roof_aperture_area()

    def add_mechanical_collection(self, _mech_collection: PhxMechanicalSystemCollection) -> None:
        """Add a new mechanical collection to the variant."""
        self._mech_collections.append(_mech_collection)

    def clear_mechanical_collections(self) -> None:
        """Clear all mechanical collections from the variant."""
        self._mech_collections = []

    def get_mech_device_by_key(
        self, _key: str
    ) -> Tuple[Optional[PhxMechanicalSystemCollection], Optional[PhxMechanicalDevice]]:
        """Return a Tuple of a mech-collection and a mechanical device based on the specified device key, or None if not found."""

        found: List[Tuple[Optional[PhxMechanicalSystemCollection], Optional[PhxMechanicalDevice]]] = []
        for mech_collection in self._mech_collections:
            found.append((mech_collection, mech_collection._devices.get(_key, None)))

        if len(found) == 0:
            return None, None
        elif len(found) == 1:
            return found[0]
        else:
            raise ValueError(f"Multiple mechanical devices found with key: {_key}.")

    def device_in_collections(self, _key: str) -> bool:
        """See if the variant's mechanical device collections already includes the specified key."""
        for mech_collection in self._mech_collections:
            if mech_collection.device_in_collection(_key):
                return True
        return False

    def supportive_device_in_collections(self, _key: str) -> bool:
        """Return a supportive device based on the specified device key, or None if not found."""
        for mech_collection in self._mech_collections:
            if mech_collection.supportive_devices.device_in_collection(_key) == True:
                return True
        return False

    def renewable_device_in_collections(self, _key: str) -> bool:
        """Return a renewable device based on the specified device key, or None if not found."""
        for mech_collection in self._mech_collections:
            if mech_collection.renewable_devices.device_in_collection(_key) == True:
                return True
        return False

    def get_mech_device_by_id(self, _id_num: int) -> PhxMechanicalDevice:
        """Returns a Mechanical Device from the collections which has a matching id-num."""
        for mech_collection in self._mech_collections:
            phx_mech_ventilator = mech_collection.get_mech_device_by_id(_id_num)
            if phx_mech_ventilator:
                return phx_mech_ventilator

        raise NoDeviceFoundError(_id_num)


@dataclass
class ProjectData_Agent:
    name: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    post_code: Optional[str] = None
    telephone: Optional[str] = None
    email: Optional[str] = None
    license_number: Optional[str] = None


@dataclass
class PhxProjectDate:
    year: int = datetime.now().year
    month: int = datetime.now().month
    day: int = datetime.now().day
    hour: int = datetime.now().hour
    minutes: int = datetime.now().minute


@dataclass
class PhxProjectData:
    customer: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    building: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    owner: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    designer: ProjectData_Agent = field(default_factory=ProjectData_Agent)

    project_date: PhxProjectDate = field(default_factory=PhxProjectDate)
    owner_is_client: bool = False
    year_constructed: int = 0
    image: Optional[bool] = None


@dataclass
class PhxProject:
    name: str = "unnamed_project"
    assembly_types: Dict[str, PhxConstructionOpaque] = field(default_factory=dict)
    window_types: Dict[str, PhxConstructionWindow] = field(default_factory=dict)
    shade_types: Dict[str, PhxWindowShade] = field(default_factory=dict)
    utilization_patterns_ventilation: UtilizationPatternCollection_Ventilation = field(
        default_factory=UtilizationPatternCollection_Ventilation
    )
    utilization_patterns_occupancy: UtilizationPatternCollection_Occupancy = field(
        default_factory=UtilizationPatternCollection_Occupancy
    )
    utilization_patterns_lighting: UtilizationPatternCollection_Lighting = field(
        default_factory=UtilizationPatternCollection_Lighting
    )
    variants: List[PhxVariant] = field(default_factory=list)
    project_data: PhxProjectData = field(default_factory=PhxProjectData)
    data_version: int = 48
    unit_system: int = 1  # SI
    program_version: str = "3.2.0.1"
    scope: int = 3
    visualized_geometry: int = 2

    def add_new_variant(self, _variant: PhxVariant) -> None:
        """Adds a new PHX Variant to the Project."""
        self.variants.append(_variant)

    @property
    def assembly_type_id_numbers(self) -> Set[int]:
        return {assembly.id_num for assembly in self.assembly_types.values()}

    @property
    def window_type_id_numbers(self) -> Set[int]:
        return {window.id_num for window in self.window_types.values()}

    @property
    def shade_type_id_numbers(self) -> Set[int]:
        return {shade.id_num for shade in self.shade_types.values()}

    def add_assembly_type(self, _assembly_type: PhxConstructionOpaque, _key=None) -> None:
        """Adds a new PhxConstructionOpaque to the Project's collection"""
        # -- be sure to the ID-Num
        if _assembly_type.id_num in self.assembly_type_id_numbers:
            _assembly_type.id_num = max(self.assembly_type_id_numbers) + 1

        if _key:
            self.assembly_types[_key] = _assembly_type
        else:
            self.assembly_types[_assembly_type.identifier] = _assembly_type

    def add_new_window_type(self, _window_type: PhxConstructionWindow, _key=None) -> None:
        """Adds a new PhxConstructionWindow to the Project's collection"""
        if _window_type.id_num in self.window_type_id_numbers:
            _window_type.id_num = max(self.window_type_id_numbers) + 1

        if _key:
            self.window_types[_key] = _window_type
        else:
            self.window_types[_window_type.identifier] = _window_type

    def get_window_type(self, _key: str) -> PhxConstructionWindow:
        """Returns the PhxConstructionWindow with the specified key"""
        try:
            return self.window_types[_key]
        except KeyError as e:
            valid_keys = "  |  ".join([f"{k}::{v.display_name}" for k, v in self.window_types.items()])
            msg = (
                f"Window Type: '{_key}' not found in project collection? "
                f"Valid window-types include only: {valid_keys}."
            )
            raise KeyError(msg)

    def get_window_types_by_name(self, _name: str) -> List[PhxConstructionWindow]:
        """Returns a list of PhxConstructionWindow with the specified name.

        CAUTION: Multiple window types can have the same name.
        Use the .get_window_type() and use the 'key' to get an exact window type.
        """
        return [window_type for window_type in self.window_types.values() if window_type.display_name == _name]

    def add_new_shade_type(self, _shade_type: PhxWindowShade, _key=None) -> None:
        """Adds a new PhxWindowShade to the Project's collection"""
        if _shade_type.id_num in self.shade_type_id_numbers:
            _shade_type.id_num = max(self.shade_type_id_numbers) + 1

        if _key:
            self.shade_types[_key] = _shade_type
        else:
            self.shade_types[_shade_type.identifier] = _shade_type

    def vent_sched_in_project_collection(self, _key: str) -> bool:
        """See if the project Ventilation schedule collection already includes the specified key."""
        return self.utilization_patterns_ventilation.key_is_in_collection(_key)

    def occupancy_sched_in_project_collection(self, _key: str) -> bool:
        """See if the project Occupancy schedule collection already includes the specified key."""
        return self.utilization_patterns_occupancy.key_is_in_collection(_key)

    def lighting_sched_in_project_collection(self, _key: str) -> bool:
        """See if the project Lighting schedule collection already includes the specified key."""
        return self.utilization_patterns_lighting.key_is_in_collection(_key)

    def add_vent_sched_to_collection(self, vent_sched: ventilation.PhxScheduleVentilation) -> None:
        """Add a new Ventilation schedule to the project's collection."""
        self.utilization_patterns_ventilation.add_new_util_pattern(vent_sched)

    def add_occupancy_sched_to_collection(self, vent_sched: Optional[occupancy.PhxScheduleOccupancy]) -> None:
        """Add a new Occupancy schedule to the project's collection."""
        self.utilization_patterns_occupancy.add_new_util_pattern(vent_sched)

    def add_lighting_sched_to_collection(self, lighting_sched: Optional[lighting.PhxScheduleLighting]) -> None:
        """Add a new Occupancy schedule to the project's collection."""
        self.utilization_patterns_lighting.add_new_util_pattern(lighting_sched)

    def get_total_gross_wall_area(self) -> float:
        """Get the total gross wall area for all variants in the project (ignoring any apertures)."""
        return sum([variant.get_total_gross_wall_area() for variant in self.variants])

    def get_total_net_wall_area(self) -> float:
        """Get the total net wall area for all variants in the project (ignoring any apertures)."""
        return sum([variant.get_total_net_wall_area() for variant in self.variants])

    def get_total_gross_roof_area(self) -> float:
        """Get the total gross roof area for all variants in the project."""
        return sum([variant.get_total_gross_roof_area() for variant in self.variants])

    def get_total_net_roof_area(self) -> float:
        """Get the total net roof area for all variants in the project."""
        return sum([variant.get_total_net_roof_area() for variant in self.variants])

    def get_total_wall_aperture_area(self) -> float:
        """Get the total window area for all variants in the project."""
        return sum([variant.get_total_wall_aperture_area() for variant in self.variants])

    def get_total_roof_aperture_area(self) -> float:
        """Get the total roof aperture area for all variants in the project."""
        return sum([variant.get_total_roof_aperture_area() for variant in self.variants])

    def __str__(self):
        return f"{self.__class__.__name__}"



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Valid 'types' for Building Elements."""

from enum import Enum


class ElectricEquipmentType(Enum):
    DISHWASHER = 1
    CLOTHES_WASHER = 2
    CLOTHES_DRYER = 3
    REFRIGERATOR = 4
    FREEZER = 5
    FRIDGE_FREEZER = 6
    COOKING = 7
    CUSTOM = 11
    MEL = 13
    LIGHTING_INTERIOR = 14
    LIGHTING_EXTERIOR = 15
    LIGHTING_GARAGE = 16
    CUSTOM_LIGHTING = 17
    CUSTOM_MEL = 18



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP "SolarDHW" worksheet."""

from __future__ import annotations

from dataclasses import dataclass, field
from functools import partial
from typing import Any, Dict

from ph_units.unit_type import Unit

from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_app


@dataclass
class SolarDHWData:
    """Convenience class for organizing and cleaning the data."""

    footprint: Unit = field(default_factory=Unit)
    annual_dhw_energy: Unit = field(default_factory=Unit)
    annual_dhw_contribution: Unit = field(default_factory=Unit)
    annual_heating_energy: Unit = field(default_factory=Unit)
    annual_heating_contribution: Unit = field(default_factory=Unit)

    @classmethod
    def from_PHPP_data(cls, _data: Dict[str, Any]) -> SolarDHWData:
        """Clean up the data coming in from PHPP"""
        obj = cls()

        obj.footprint = Unit(_data["footprint"], _data["footprint_unit"])
        obj.annual_dhw_energy = Unit(_data["dhw_energy"], _data["energy_unit"])
        obj.annual_dhw_contribution = Unit(_data["dhw_contribution"], "-")
        obj.annual_heating_energy = Unit(_data["heating_energy"], _data["energy_unit"])
        obj.annual_heating_contribution = Unit(_data["heating_contribution"], "-")

        return obj


class SolarDHW:
    """IO Controller for the PHPP 'Solar DHW' worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.SolarDhw) -> None:
        self.xl = _xl
        self.shape = _shape

    def get_phpp_data(self) -> SolarDHWData:
        """Get the data from the PHPP worksheet."""
        get_ = partial(self.xl.get_single_data_item, self.shape.name)
        ranges = self.shape.ranges
        return SolarDHWData.from_PHPP_data(
            {
                "footprint_unit": self.shape.footprint_unit,
                "energy_unit": self.shape.energy_unit,
                "footprint": get_(ranges.footprint),
                "dhw_energy": get_(ranges.annual_dhw_energy),
                "dhw_contribution": get_(ranges.annual_dhw_contribution),
                "heating_energy": get_(ranges.annual_heating_energy),
                "heating_contribution": get_(ranges.annual_heating_contribution),
            }
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP "Cooling Units" worksheet."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from ph_units.unit_type import Unit

from PHX.model.enums.hvac import CoolingType
from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_app


@dataclass
class CoolingUnitData:
    """Convenience class for organizing and cleaning the data."""

    used: bool = False
    SEER: Unit = field(default_factory=Unit)
    num_units: int = 1
    _device_type_name: Optional[str] = None
    device_type: CoolingType = field(default=CoolingType.NONE)

    @property
    def device_type_name(self) -> Optional[str]:
        return self._device_type_name

    @device_type_name.setter
    def device_type_name(self, _value: Optional[str]) -> None:
        if _value is None:
            self._device_type_name = None
            return
        self._device_type_name = str(_value).split("-", 1)[-1].strip()

    @classmethod
    def from_PHPP_data(cls, _data: Dict[str, Any], _seer_unit: str) -> CoolingUnitData:
        """Clean up the data coming in from PHPP"""
        obj = cls()

        if _data.get("used", None) in ["X", "x"]:
            obj.used = True
        else:
            return obj

        obj.SEER = Unit(_data.get("SEER", 0.0), _seer_unit)
        obj.num_units = int(float(_data.get("num_units", 1.0)))
        # -- Since using named ranges, brings in a list of the merged columns...
        # -- Only want the first one.
        obj.device_type_name = str(_data.get("device_type_name", [None])[0])
        obj.device_type = _data["device_type"]

        return obj


class SupplyAir:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.CoolingUnits) -> None:
        self.xl = _xl
        self.shape = _shape

    def get_phpp_data(self) -> CoolingUnitData:
        shape = self.shape.supply_air
        data = {
            "used": self.xl.get_data(self.shape.name, shape.used),
            "SEER": self.xl.get_data(self.shape.name, shape.SEER),
            "num_units": self.xl.get_data(self.shape.name, shape.num_units),
            "device_type_name": self.xl.get_data(self.shape.name, shape.device_type_name),
            "device_type": CoolingType.VENTILATION,
        }
        return CoolingUnitData.from_PHPP_data(data, self.shape.SEER_unit)


class RecirculationAir:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.CoolingUnits) -> None:
        self.xl = _xl
        self.shape = _shape

    def get_phpp_data(self) -> CoolingUnitData:
        shape = self.shape.recirculation_air
        data = {
            "used": self.xl.get_data(self.shape.name, shape.used),
            "SEER": self.xl.get_data(self.shape.name, shape.SEER),
            "num_units": self.xl.get_data(self.shape.name, shape.num_units),
            "device_type_name": self.xl.get_data(self.shape.name, shape.device_type_name),
            "device_type": CoolingType.RECIRCULATION,
        }
        return CoolingUnitData.from_PHPP_data(data, self.shape.SEER_unit)


class Dehumidification:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.CoolingUnits) -> None:
        self.xl = _xl
        self.shape = _shape

    def get_phpp_data(self) -> CoolingUnitData:
        shape = self.shape.dehumidification
        data = {
            "used": self.xl.get_data(self.shape.name, shape.used),
            "SEER": self.xl.get_data(self.shape.name, shape.SEER),
            "device_type": CoolingType.DEHUMIDIFICATION,
        }
        return CoolingUnitData.from_PHPP_data(data, self.shape.SEER_unit)


class Panel:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.CoolingUnits) -> None:
        self.xl = _xl
        self.shape = _shape

    def get_phpp_data(self) -> CoolingUnitData:
        shape = self.shape.panel
        data = {
            "used": self.xl.get_data(self.shape.name, shape.used),
            "SEER": self.xl.get_data(self.shape.name, shape.SEER),
            "device_type_name": self.xl.get_data(self.shape.name, shape.device_type_name),
            "device_type": CoolingType.PANEL,
        }
        return CoolingUnitData.from_PHPP_data(data, self.shape.SEER_unit)


class CoolingUnits:
    """IO Controller for the PHPP Cooling Units worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.CoolingUnits) -> None:
        self.xl = _xl
        self.shape = _shape
        self.supply_air = SupplyAir(self.xl, self.shape)
        self.recirculation_air = RecirculationAir(self.xl, self.shape)
        self.dehumidification = Dehumidification(self.xl, self.shape)
        self.panel = Panel(self.xl, self.shape)

    def get_cooling_system_data(self) -> Tuple[CoolingUnitData, ...]:
        return (
            self.supply_air.get_phpp_data(),
            self.recirculation_air.get_phpp_data(),
            self.dehumidification.get_phpp_data(),
            self.panel.get_phpp_data(),
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Classes for converting WUFI-Pydantic Entities to PHX Model Objects."""

import sys
from functools import partial
from typing import Any, Dict, List, Optional, Tuple

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.polyline import LineSegment3D
from ph_units.converter import convert
from rich import print

from PHX.from_WUFI_XML import wufi_file_schema as wufi_xml
from PHX.model.building import PhxBuilding, PhxZone
from PHX.model.certification import PhxPhBuildingData, PhxPhiusCertification
from PHX.model.components import PhxApertureElement, PhxComponentAperture, PhxComponentOpaque, PhxComponentThermalBridge
from PHX.model.constructions import (
    PhxConstructionOpaque,
    PhxConstructionWindow,
    PhxLayer,
    PhxMaterial,
    PhxWindowFrameElement,
)
from PHX.model.elec_equip import (
    PhxDeviceClothesDryer,
    PhxDeviceClothesWasher,
    PhxDeviceCooktop,
    PhxDeviceCustomElec,
    PhxDeviceCustomLighting,
    PhxDeviceCustomMEL,
    PhxDeviceDishwasher,
    PhxDeviceFreezer,
    PhxDeviceFridgeFreezer,
    PhxDeviceLightingExterior,
    PhxDeviceLightingGarage,
    PhxDeviceLightingInterior,
    PhxDeviceMEL,
    PhxDeviceRefrigerator,
    PhxElectricalDevice,
)
from PHX.model.enums import hvac as hvac_enums
from PHX.model.enums.building import (
    AttachedZoneType,
    ComponentColor,
    ComponentExposureExterior,
    ComponentFaceOpacity,
    ComponentFaceType,
    SpecificHeatCapacity,
    ThermalBridgeType,
    WindExposureType,
    ZoneType,
)
from PHX.model.enums.elec_equip import ElectricEquipmentType
from PHX.model.enums.foundations import FoundationType
from PHX.model.enums.phius_certification import (
    PhiusCertificationBuildingCategoryType,
    PhiusCertificationBuildingStatus,
    PhiusCertificationBuildingType,
    PhiusCertificationBuildingUseType,
    PhiusCertificationProgram,
)
from PHX.model.enums.phx_site import SiteClimateSelection, SiteEnergyFactorSelection, SiteSelection
from PHX.model.geometry import PhxPlane, PhxPolygon, PhxVector, PhxVertix
from PHX.model.ground import PhxFoundation, PhxHeatedBasement, PhxSlabOnGrade, PhxUnHeatedBasement, PhxVentedCrawlspace
from PHX.model.hvac import (
    AnyPhxHeaterBoiler,
    PhxHeaterBoilerFossil,
    PhxHeaterBoilerWood,
    PhxHeaterDistrictHeat,
    PhxHeaterElectric,
    PhxHeatPumpAnnual,
    PhxHeatPumpCombined,
    PhxHeatPumpDevice,
    PhxHeatPumpHotWater,
    PhxHeatPumpMonthly,
)
from PHX.model.hvac._base import PhxUsageProfile
from PHX.model.hvac.collection import (
    AnyMechDevice,
    PhxMechanicalSystemCollection,
    PhxRecirculationParameters,
    PhxZoneCoverage,
)
from PHX.model.hvac.cooling_params import (
    PhxCoolingDehumidificationParams,
    PhxCoolingPanelParams,
    PhxCoolingRecirculationParams,
    PhxCoolingVentilationParams,
)
from PHX.model.hvac.ducting import PhxDuctElement, PhxDuctSegment, PhxVentDuctType
from PHX.model.hvac.piping import (
    PhxHotWaterPipingCalcMethod,
    PhxHotWaterPipingInchDiameterType,
    PhxHotWaterPipingMaterial,
    PhxHotWaterSelectionUnitsOrFloors,
    PhxPipeBranch,
    PhxPipeElement,
    PhxPipeSegment,
    PhxPipeTrunk,
)
from PHX.model.hvac.renewable_devices import PhxDevicePhotovoltaic
from PHX.model.hvac.supportive_devices import PhxSupportiveDevice, PhxSupportiveDeviceType
from PHX.model.hvac.ventilation import (
    AnyPhxExhaustVent,
    PhxDeviceVentilator,
    PhxExhaustVentilatorDryer,
    PhxExhaustVentilatorRangeHood,
    PhxExhaustVentilatorUserDefined,
)
from PHX.model.hvac.water import PhxHotWaterTank
from PHX.model.phx_site import PhxClimate, PhxCO2Factor, PhxGround, PhxPEFactor, PhxSite, PhxSiteEnergyFactors
from PHX.model.project import PhxProject, PhxProjectData, PhxProjectDate, PhxVariant, ProjectData_Agent, WufiPlugin
from PHX.model.schedules.lighting import PhxScheduleLighting
from PHX.model.schedules.occupancy import PhxScheduleOccupancy
from PHX.model.schedules.ventilation import PhxScheduleVentilation
from PHX.model.shades import PhxWindowShade
from PHX.model.spaces import PhxSpace

# -----------------------------------------------------------------------------
# -- Conversion Function


def as_phx_obj(_model, _schema_name, **kwargs) -> Any:
    """Find the right class-builder from the module and pass along the data to it.

    Args:
        _model (Any): The Pydantic (BaseModel) item to be converted.
        _schema_name ([type]): The name of the class-builder to be used.
        kwargs: Any additional arguments to be passed to the class-builder.
    Returns:
        A new PHX object built from the input data
    """
    builder = getattr(sys.modules[__name__], f"_{_schema_name}")
    return builder(_model, **kwargs)


# -----------------------------------------------------------------------------
# -- Project


def _PhxProject(_model: wufi_xml.WUFIplusProject) -> PhxProject:
    phx_obj = PhxProject()
    phx_obj.data_version = _model.DataVersion
    phx_obj.unit_system = _model.UnitSystem
    phx_obj.program_version = _model.ProgramVersion
    phx_obj.scope = _model.Scope
    phx_obj.visualized_geometry = _model.DimensionsVisualizedGeometry

    phx_obj.project_data = as_phx_obj(_model.ProjectData, "PhxProjectData")

    # ----------------------------------------------------------------------
    # -- Build all the type collections first
    for window_type_data in _model.WindowTypes or []:
        new_window = as_phx_obj(window_type_data, "PhxConstructionWindow")
        # -- Be sure to use the identifier as the key so the Component
        # -- lookup works properly. We don't use the name here since
        # -- wufi doesn't enforce unique names like HB does.
        phx_obj.add_new_window_type(new_window, _key=new_window.identifier)

    for assembly_data in _model.Assemblies or []:
        new_opaque: PhxConstructionOpaque = as_phx_obj(assembly_data, "PhxConstructionOpaque")
        # -- Be sure to use the identifier as the key so the Component
        # -- lookup works properly. We don't use the name here since
        # -- wufi doesn't enforce unique names like HB does.
        phx_obj.add_assembly_type(new_opaque, _key=new_opaque.identifier)

    for shade_type_data in _model.SolarProtectionTypes or []:
        new_shade: PhxWindowShade = as_phx_obj(shade_type_data, "PhxWindowShade")
        # -- Be sure to use the identifier as the key so the Component
        # -- lookup works properly. We don't use the name here since
        # -- wufi doesn't enforce unique names like HB does.
        phx_obj.add_new_shade_type(new_shade, _key=new_shade.identifier)

    for vent_pattern_data in _model.UtilisationPatternsVentilation or []:
        new_pattern = as_phx_obj(vent_pattern_data, "PhxScheduleVentilation")
        phx_obj.utilization_patterns_ventilation.add_new_util_pattern(new_pattern)

    for ph_pattern_data in _model.UtilizationPatternsPH or []:
        new_pattern = as_phx_obj(ph_pattern_data, "PhxScheduleOccupancy")
        phx_obj.utilization_patterns_occupancy.add_new_util_pattern(new_pattern)

    # ----------------------------------------------------------------------
    # -- Build all the actual Variants
    for variant_dict in _model.Variants or []:
        new_variant = as_phx_obj(variant_dict, "PhxVariant", _phx_project_host=phx_obj)
        phx_obj.add_new_variant(new_variant)

    return phx_obj


def _PhxProjectDate(_date: wufi_xml.WufiDateProject) -> PhxProjectDate:
    phx_obj = PhxProjectDate()
    if _date:
        phx_obj.year = _date.Year
        phx_obj.month = _date.Month
        phx_obj.day = _date.Day
        phx_obj.hour = _date.Hour
        phx_obj.minutes = _date.Minutes
    return phx_obj


def _PhxProjectData(_model: wufi_xml.WufiProjectData) -> PhxProjectData:
    phx_obj = PhxProjectData()
    phx_obj.project_date = as_phx_obj(_model.Date_Project, "PhxProjectDate")
    phx_obj.owner_is_client = _model.OwnerIsClient or False
    phx_obj.year_constructed = int(_model.Year_Construction or 0)
    phx_obj.image = _model.WhiteBackgroundPictureBuilding

    phx_obj.customer = ProjectData_Agent(
        _model.Customer_Name,
        _model.Customer_Street,
        _model.Customer_Locality,
        _model.Customer_PostalCode,
        _model.Customer_Tel,
        _model.Customer_Email,
    )
    phx_obj.building = ProjectData_Agent(
        _model.Building_Name,
        _model.Building_Street,
        _model.Building_Locality,
        _model.Building_PostalCode,
    )
    phx_obj.owner = ProjectData_Agent(
        _model.Owner_Name,
        _model.Owner_Street,
        _model.Owner_Locality,
        _model.Owner_PostalCode,
    )
    phx_obj.designer = ProjectData_Agent(
        _model.Responsible_Name,
        _model.Responsible_Street,
        _model.Responsible_Locality,
        _model.Responsible_PostalCode,
        _model.Responsible_Tel,
        _model.Responsible_LicenseNr,
        _model.Responsible_Email,
    )

    return phx_obj


# -----------------------------------------------------------------------------
# -- Envelope Types


def _PhxConstructionWindow(_t: wufi_xml.WufiWindowType) -> PhxConstructionWindow:
    phx_obj = PhxConstructionWindow()
    phx_obj.id_num = _t.IdentNr
    phx_obj.identifier = str(_t.IdentNr)
    phx_obj.display_name = _t.Name
    phx_obj.use_detailed_uw = _t.Uw_Detailed
    phx_obj.use_detailed_frame = _t.GlazingFrameDetailed

    phx_obj.u_value_window = _t.U_Value
    phx_obj.u_value_glass = _t.U_Value_Glazing
    phx_obj.u_value_frame = _t.U_Value_Frame

    phx_obj.glass_mean_emissivity = _t.MeanEmissivity
    phx_obj.glass_g_value = _t.g_Value

    # -- if there is no value provide, fall back to the last value found...
    # -- I *think* there will always be a 'left' frame-element?
    frame_data_left = {
        "width": _t.Frame_Width_Left or 0.1,
        "u_value": _t.Frame_U_Left or 1.0,
        "psi_glazing": _t.Glazing_Psi_Left or 0.0,
        "psi_install": _t.Frame_Psi_Left or 0.0,
    }
    phx_obj.frame_left = PhxWindowFrameElement(**frame_data_left)

    frame_data_right = {
        "width": _t.Frame_Width_Right or frame_data_left["width"],
        "u_value": _t.Frame_U_Right or frame_data_left["u_value"],
        "psi_glazing": _t.Glazing_Psi_Right or frame_data_left["psi_glazing"],
        "psi_install": _t.Frame_Psi_Right or frame_data_left["psi_install"],
    }
    phx_obj.frame_right = PhxWindowFrameElement(**frame_data_right)

    frame_data_top = {
        "width": _t.Frame_Width_Top or frame_data_right["width"],
        "u_value": _t.Frame_U_Top or frame_data_right["u_value"],
        "psi_glazing": _t.Glazing_Psi_Top or frame_data_right["psi_glazing"],
        "psi_install": _t.Frame_Psi_Top or frame_data_right["psi_install"],
    }
    phx_obj.frame_top = PhxWindowFrameElement(**frame_data_top)

    frame_data_bottom = {
        "width": _t.Frame_Width_Top or frame_data_top["width"],
        "u_value": _t.Frame_U_Top or frame_data_top["u_value"],
        "psi_glazing": _t.Glazing_Psi_Top or frame_data_top["psi_glazing"],
        "psi_install": _t.Frame_Psi_Top or frame_data_top["psi_install"],
    }
    phx_obj.frame_bottom = PhxWindowFrameElement(**frame_data_bottom)

    return phx_obj


def _PhxConstructionOpaque(_data: wufi_xml.WufiAssembly) -> PhxConstructionOpaque:
    phx_obj = PhxConstructionOpaque()
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.display_name = _data.Name
    phx_obj.layer_order = _data.Order_Layers
    phx_obj.grid_kind = _data.Grid_Kind
    for layer in _data.Layers:
        new_layer = as_phx_obj(layer, "PhxLayer")
        phx_obj.layers.append(new_layer)

    return phx_obj


def _PhxLayer(_data: wufi_xml.WufiLayer) -> PhxLayer:
    phx_obj = PhxLayer()
    phx_obj.thickness_m = _data.Thickness
    new_mat = as_phx_obj(_data.Material, "PhxMaterial")
    phx_obj.set_material(new_mat)

    return phx_obj


def _PhxMaterial(_data: wufi_xml.WufiMaterial) -> PhxMaterial:
    phx_obj = PhxMaterial()
    phx_obj.display_name = _data.Name
    phx_obj.conductivity = _data.ThermalConductivity
    phx_obj.density = _data.BulkDensity
    phx_obj.porosity = _data.Porosity
    phx_obj.heat_capacity = _data.HeatCapacity
    phx_obj.water_vapor_resistance = _data.WaterVaporResistance
    phx_obj.reference_water = _data.ReferenceWaterContent
    return phx_obj


def _PhxWindowShade(_data: wufi_xml.WufiSolarProtectionType) -> PhxWindowShade:
    phx_obj = PhxWindowShade()
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.display_name = _data.Name
    phx_obj.operation_mode = _data.OperationMode
    phx_obj.reduction_factor = _data.MaxRedFactorRadiation or 1.0
    return phx_obj


# -----------------------------------------------------------------------------
# -- Utilization Patterns


def _PhxScheduleVentilation(
    _data: wufi_xml.WufiUtilizationPatternVent,
) -> PhxScheduleVentilation:
    phx_obj = PhxScheduleVentilation()

    phx_obj.id_num = _data.IdentNr
    phx_obj.name = _data.Name
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.operating_days = _data.OperatingDays
    phx_obj.operating_weeks = _data.OperatingWeeks
    phx_obj.operating_periods.high.period_operating_hours = _data.Maximum_DOS
    phx_obj.operating_periods.high.period_operation_speed = _data.Maximum_PDF
    phx_obj.operating_periods.standard.period_operating_hours = _data.Standard_DOS
    phx_obj.operating_periods.standard.period_operation_speed = _data.Standard_PDF
    phx_obj.operating_periods.basic.period_operating_hours = _data.Basic_DOS
    phx_obj.operating_periods.basic.period_operation_speed = _data.Basic_PDF
    phx_obj.operating_periods.minimum.period_operating_hours = _data.Minimum_DOS
    phx_obj.operating_periods.minimum.period_operation_speed = _data.Minimum_PDF
    return phx_obj


def _PhxScheduleOccupancy(_data: wufi_xml.WufiUtilizationPattern) -> PhxScheduleOccupancy:
    phx_obj = PhxScheduleOccupancy()
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)
    phx_obj.display_name = _data.Name
    phx_obj.start_hour = _data.BeginUtilization
    phx_obj.end_hour = _data.EndUtilization
    phx_obj.annual_utilization_days = _data.AnnualUtilizationDays
    phx_obj.relative_utilization_factor = _data.RelativeAbsenteeism

    return phx_obj


# -----------------------------------------------------------------------------
# -- Variants


def _WufiPlugin(_model: wufi_xml.WufiPlugin) -> WufiPlugin:
    phx_obj = WufiPlugin()
    if not _model:
        return phx_obj

    phx_obj.insert_plugin = bool(_model.InsertPlugIn)
    phx_obj.name_dll = _model.Name_dll
    phx_obj.status_plugin = _model.StatusPlugIn
    return phx_obj


def _PhxVariant(_xml_variant_data: wufi_xml.WufiVariant, _phx_project_host: PhxProject) -> PhxVariant:
    phx_obj = PhxVariant()

    phx_obj.id_num = _xml_variant_data.IdentNr
    phx_obj.name = _xml_variant_data.Name
    phx_obj.remarks = _xml_variant_data.Remarks
    phx_obj.plugin = as_phx_obj(_xml_variant_data.PlugIn, "WufiPlugin")
    phx_obj.phius_cert = as_phx_obj(_xml_variant_data.PassivehouseData, "PhxPhiusCertification")
    phx_obj.site = as_phx_obj(_xml_variant_data.ClimateLocation, "PhxSite")

    phx_obj.building = as_phx_obj(
        (_xml_variant_data.Building, _xml_variant_data.Graphics_3D),
        "PhxBuilding",
        _phx_project_host=_phx_project_host,
    )

    # -- Build the HVAC Systems, Devices, and Distribution
    phx_obj.clear_mechanical_collections()
    for xml_system_data in _xml_variant_data.HVAC.Systems:
        new_mechanical_collection = PhxMechanicalSystemCollection()
        new_mechanical_collection.display_name = xml_system_data.Name
        new_mechanical_collection.id_num = xml_system_data.IdentNr
        new_mechanical_collection.zone_coverage = as_phx_obj(xml_system_data.ZonesCoverage[0], "PhxZoneCoverage")

        # ---------------------------------------------------------------------
        # -- Build all the actual Mechanical Devices (boilers, heat-pumps, etc...)
        for xml_device_data in xml_system_data.Devices:
            new_device = as_phx_obj(xml_device_data, "PhxMechanicalDevice")  # type: AnyMechDevice
            new_mechanical_collection.add_new_mech_device(new_device.identifier, new_device)

        # ---------------------------------------------------------------------
        # -- Cooling Distribution
        xml_dist_cooling_data = xml_system_data.PHDistribution.DistributionCooling
        if xml_dist_cooling_data is not None:
            # -- This part is super stupid. Since WUFI doesn't store this
            # -- device information properly (ie: ON the device) but instead over
            # -- in the 'Distribution' part as single total values, we'll just
            # -- have to make some shit up. How about we evenly distribute the
            # -- total capacity across all the cooling devices? Seems like as
            # -- good a solution as any I guess....

            all_cooling_devices = new_mechanical_collection.cooling_devices
            number_phx_cooling_devices = len(all_cooling_devices)
            if number_phx_cooling_devices == 0:
                continue
            as_phx_w_num_devices = partial(as_phx_obj, number_phx_cooling_devices=number_phx_cooling_devices)

            # -- Apply the param values to all of the cooling heat-pumps found
            for cooling_device in all_cooling_devices:
                cooling_device.params_cooling.ventilation = as_phx_w_num_devices(
                    xml_dist_cooling_data, "PhxCoolingVentilationParams"
                )
                cooling_device.params_cooling.recirculation = as_phx_w_num_devices(
                    xml_dist_cooling_data, "PhxCoolingRecirculationParams"
                )
                cooling_device.params_cooling.dehumidification = as_phx_w_num_devices(
                    xml_dist_cooling_data, "PhxCoolingDehumidificationParams"
                )
                cooling_device.params_cooling.panel = as_phx_w_num_devices(
                    xml_dist_cooling_data, "PhxCoolingPanelParams"
                )

        # ---------------------------------------------------------------------
        # -- DHW Distribution (Piping)
        xml_dist_dhw_data = xml_system_data.PHDistribution.DistributionDHW
        if xml_dist_dhw_data is not None:
            # -- Distribution Piping Elements
            for trunc in xml_dist_dhw_data.Truncs or []:
                new_trunc = as_phx_obj(trunc, "Trunc")  # type: PhxPipeTrunk
                new_mechanical_collection.add_distribution_piping(new_trunc)

            # -- Recirculation Piping Parameters
            phx_params = as_phx_obj(xml_dist_dhw_data, "PhxRecirculationParameters")
            new_mechanical_collection._distribution_hw_recirculation_params = phx_params

            # -- Recirculation Piping Element
            recirc_pipe_element = as_phx_obj(xml_dist_dhw_data, "RecirculationTrunk")  # type: PhxPipeElement
            new_mechanical_collection.add_recirc_piping(recirc_pipe_element)

        # ---------------------------------------------------------------------
        # -- Ventilation Distribution (Ducting)
        for wufi_duct_data in xml_system_data.PHDistribution.DistributionVentilation or []:
            # -- PHX has a 1:1 duct-to-ventilator relationship, so we'll just
            # -- duplicate the duct for each one its assigned to in WUFI-XML
            for i in wufi_duct_data.AssignedVentUnits or []:
                new_duct = as_phx_obj(wufi_duct_data, "PhxDuctElement", ventilator=i)
                new_mechanical_collection.add_vent_ducting(new_duct)

        # ---------------------------------------------------------------------
        # -- Supportive Devices (Pumps, Fans, etc...)
        xml_supp_device_data = xml_system_data.PHDistribution.SupportiveDevices
        for supportive_device_data in xml_supp_device_data or []:
            new_supp_device = as_phx_obj(supportive_device_data, "PhxSupportiveDevice")
            new_mechanical_collection.supportive_devices.add_new_device(new_supp_device.identifier, new_supp_device)

        phx_obj.add_mechanical_collection(new_mechanical_collection)

    return phx_obj


# -----------------------------------------------------------------------------
# -- Distribution Supportive Devices


def _PhxSupportiveDevice(_data: wufi_xml.WufiSupportiveDevice) -> PhxSupportiveDevice:
    new_supportive_device = PhxSupportiveDevice()

    new_supportive_device.display_name = _data.Name or ""
    new_supportive_device.device_type = PhxSupportiveDeviceType(_data.Type or 10)
    new_supportive_device.quantity = _data.Quantity or 0
    new_supportive_device.params.in_conditioned_space = _data.InConditionedSpace or False
    new_supportive_device.params.norm_energy_demand_W = _data.NormEnergyDemand or 0.0
    new_supportive_device.params.annual_period_operation_khrs = _data.PeriodOperation or 0.0

    return new_supportive_device


# -----------------------------------------------------------------------------
# -- Ventilation Distribution


def _PhxDuctSegment(
    _data: wufi_xml.WufiDuct,
) -> PhxDuctSegment:
    # -- WUFI doesn't keep the actual duct geometry, so we'll just make a
    # -- new Line with the same length as the original duct.
    fake_duct_geometry = LineSegment3D.from_sdl(
        s=Point3D(0, 0, 0),
        d=Vector3D(1, 0, 0),
        length=_data.DuctLength or 0.0,
    )

    # -- Build the actual segment
    new_segment = PhxDuctSegment(
        identifier=str(_data.IdentNr),
        display_name=_data.Name or "",
        geometry=fake_duct_geometry,
        diameter_m=_data.DuctDiameter or 0.0,
        height_m=_data.DuctShapeHeight or None,
        width_m=_data.DuctShapeWidth or None,
        insulation_thickness_m=_data.InsulationThickness or 0.0,
        insulation_conductivity_wmk=_data.ThermalConductivity or 0.0,
        insulation_reflective=_data.IsReflective,
    )

    return new_segment


def _PhxDuctElement(_data: wufi_xml.WufiDuct, ventilator: wufi_xml.WufiAssignedVentUnit) -> PhxDuctElement:
    new_phx_duct = PhxDuctElement(
        identifier=str(_data.IdentNr),
        display_name=_data.Name or "",
        vent_unit_id=ventilator.IdentNrVentUnit,
    )
    new_phx_duct.duct_type = PhxVentDuctType(_data.DuctType)

    # -- Build and add the Duct Segments
    new_phx_duct.add_segment(as_phx_obj(_data, "PhxDuctSegment"))

    return new_phx_duct


# -----------------------------------------------------------------------------
# -- Cooling Distribution


def _PhxCoolingVentilationParams(
    _data: wufi_xml.WufiDistributionCooling, number_phx_cooling_devices: int = 1
) -> PhxCoolingVentilationParams:
    phx_obj = PhxCoolingVentilationParams()

    phx_obj.used = _data.CoolingViaVentilationAir or False
    phx_obj.single_speed = _data.SupplyAirCoolingOnOff or False
    phx_obj.min_coil_temp = _data.MinTemperatureCoolingCoilSupplyAir or 0.0
    phx_obj.capacity = _data.MaxSupplyAirCoolingPower or 0.0 / number_phx_cooling_devices
    phx_obj.annual_COP = _data.SupplyAirCoolinCOP or 0.0

    return phx_obj


def _PhxCoolingRecirculationParams(
    _data: wufi_xml.WufiDistributionCooling, number_phx_cooling_devices: int = 1
) -> PhxCoolingRecirculationParams:
    phx_obj = PhxCoolingRecirculationParams()

    phx_obj.used = _data.CoolingViaRecirculation or False
    phx_obj.single_speed = _data.RecirculatingAirOnOff or False
    phx_obj.min_coil_temp = _data.MinTempCoolingCoilRecirculatingAir or 0.0
    phx_obj.capacity = _data.MaxRecirculationAirCoolingPower or 0.0 / number_phx_cooling_devices
    phx_obj.annual_COP = _data.RecirculationCoolingCOP or 0.0
    phx_obj.flow_rate_m3_hr = _data.RecirculationAirVolume or 0.0 / number_phx_cooling_devices
    phx_obj.flow_rate_variable = _data.ControlledRecirculationVolumeFlow

    return phx_obj


def _PhxCoolingDehumidificationParams(
    _data: wufi_xml.WufiDistributionCooling, number_phx_cooling_devices: int = 1
) -> PhxCoolingDehumidificationParams:
    phx_obj = PhxCoolingDehumidificationParams()

    phx_obj.used = _data.Dehumidification or False
    phx_obj.annual_COP = _data.DehumdificationCOP or 0.0
    phx_obj.useful_heat_loss = _data.UsefullDehumidificationHeatLoss or False

    return phx_obj


def _PhxCoolingPanelParams(
    _data: wufi_xml.WufiDistributionCooling, number_phx_cooling_devices: int = 1
) -> PhxCoolingPanelParams:
    phx_obj = PhxCoolingPanelParams()

    phx_obj.used = _data.PanelCooling or False
    phx_obj.annual_COP = _data.DehumdificationCOP or 0.0

    return phx_obj


# -----------------------------------------------------------------------------
# -- DHW Distribution


def _PhxRecirculationParameters(
    _data: wufi_xml.WufiDistributionDHW,
) -> PhxRecirculationParameters:
    new_params = PhxRecirculationParameters()

    new_params.calc_method = PhxHotWaterPipingCalcMethod(_data.CalculationMethodIndividualPipes)
    new_params.pipe_material = PhxHotWaterPipingMaterial(_data.PipeMaterialSimplifiedMethod)
    new_params.demand_recirc = _data.DemandRecirculation
    new_params.num_bathrooms = _data.NumberOfBathrooms or 0
    new_params.all_pipes_insulated = _data.AllPipesAreInsulated
    new_params.units_or_floors = PhxHotWaterSelectionUnitsOrFloors(_data.SelectionUnitsOrFloors)
    new_params.pipe_diameter = _data.PipeDiameterSimplifiedMethod or 0.0
    new_params.air_temp = _data.TemperatureRoom_WR or 20.0
    new_params.water_temp = _data.DesignFlowTemperature_WR or 50.0
    new_params.daily_recirc_hours = _data.DailyRunningHoursCirculation_WR or 24.0

    return new_params


def _RecirculationTrunk(_data: wufi_xml.WufiDistributionDHW) -> PhxPipeElement:
    new_recirc_pipe = PhxPipeElement()

    new_segment = PhxPipeSegment.from_length(
        "Recirculation Pipe",
        _data.LengthCirculationPipes_WR or 0.0,
        PhxHotWaterPipingMaterial.COPPER_K,
        0.0254,
    )
    new_segment.insulation_thickness_m = 25.4 / 1_000
    new_segment.insulation_conductivity = 0.04

    known_heat_loss_WMK = _data.HeatLossCoefficient_WR or 0.0

    # -- Try and determine the insulation conductivity, using the
    # -- known heat-loss conductivity value as a target.
    # -- We have to do this since the heat-loss-coeff is only solved for
    # -- based on the piping when exporting to WUFI-XML from PHX.
    # -- And for some reason the actual Pipe values are not saved in the XML?
    insul_conductivity = new_segment.reverse_solve_for_insulation_conductivity(
        target_result=known_heat_loss_WMK,
        starting_conductivity=0.1,
    )
    new_segment.insulation_conductivity = insul_conductivity
    new_recirc_pipe.add_segment(new_segment)

    return new_recirc_pipe


def _Trunc(_data: wufi_xml.WufiTrunc) -> PhxPipeTrunk:
    new_trunc = PhxPipeTrunk()
    new_trunc.display_name = _data.Name or ""
    new_trunc.multiplier = _data.CountUnitsOrFloors or 1

    # -- Since we lose the actual geometry in WUFI,
    # -- we'll just make a new Pipe Element with the right length
    existing_type = PhxHotWaterPipingInchDiameterType(_data.PipingDiameter)
    existing_type_as_m_value: float = convert(existing_type.name_as_float, "IN", "M") or 0.0254
    new_trunc.pipe_element.add_segment(
        PhxPipeSegment.from_length(
            _data.Name or "",
            _data.PipingLength or 0.0,
            PhxHotWaterPipingMaterial(_data.PipeMaterial),
            existing_type_as_m_value,
        )
    )

    for branch in _data.Branches or []:
        new_branch = as_phx_obj(branch, "Branch")  # type: PhxPipeBranch
        new_trunc.add_branch(new_branch)

    return new_trunc


def _Branch(_data: wufi_xml.WufiBranch) -> PhxPipeBranch:
    new_branch = PhxPipeBranch()
    new_branch.display_name = _data.Name or ""

    # -- Since we lose the actual geometry in WUFI,
    # -- we'll just make a new Pipe Element with the right length
    existing_type = PhxHotWaterPipingInchDiameterType(_data.PipingDiameter)
    existing_type_as_m_value: float = convert(existing_type.name_as_float, "IN", "M") or 0.0254
    new_branch.pipe_element.add_segment(
        PhxPipeSegment.from_length(
            _data.Name or "",
            _data.PipingLength or 0.0,
            PhxHotWaterPipingMaterial(_data.PipeMaterial),
            existing_type_as_m_value,
        )
    )

    for pipe in _data.Twigs or []:
        new_fixture = as_phx_obj(pipe, "Fixture")  # type: PhxPipeElement
        new_branch.add_fixture(new_fixture)

    return new_branch


def _Fixture(_data: wufi_xml.WufiTwig) -> PhxPipeElement:
    new_element = PhxPipeElement()

    # -- Since we lose the actual geometry in WUFI,
    # -- we'll just make a new Pipe Element with the right length
    existing_type = PhxHotWaterPipingInchDiameterType(_data.PipingDiameter)
    existing_type_as_m_value: float = convert(existing_type.name_as_float, "IN", "M") or 0.0254
    new_element.add_segment(
        PhxPipeSegment.from_length(
            _data.Name or "",
            _data.PipingLength or 0.0,
            PhxHotWaterPipingMaterial(_data.PipeMaterial),
            existing_type_as_m_value,
        )
    )

    return new_element


# -----------------------------------------------------------------------------
# -- Building & Geometry


def _get_compo_data_by_type(
    _data: List[wufi_xml.WufiComponent],
) -> Tuple[List[wufi_xml.WufiComponent], List[wufi_xml.WufiComponent]]:
    """Helper function to sort the component dictionaries by type.

    Args:
        _data (List[Dict]): The list of component dictionaries.
    Returns:
        Tuple[List[Dict], List[Dict]]: The sorted lists of component dictionaries.
        - [0] = List of opaque component dicts
        - [1] = List of aperture components dicts
    """

    ap_dicts: List[wufi_xml.WufiComponent] = []
    opaque_dicts: List[wufi_xml.WufiComponent] = []
    for d in _data:
        if ComponentFaceOpacity(d.Type) == ComponentFaceOpacity.TRANSPARENT:
            ap_dicts.append(d)
        else:
            opaque_dicts.append(d)
    return opaque_dicts, ap_dicts


def _PhxBuilding(
    _data: Tuple[wufi_xml.WufiBuilding, wufi_xml.WufiGraphics_3D],
    _phx_project_host: PhxProject,
) -> PhxBuilding:
    phx_obj = PhxBuilding()

    # ------------------------------------------------------------------
    # -- First, build all the vertices and polygons needed by the Components
    bldg_data, geom_data = _data
    vertix_dict: Dict[int, PhxVertix] = {v.IdentNr: as_phx_obj(v, "PhxVertix") for v in geom_data.Vertices}
    polygon_dict: Dict[int, PhxPolygon] = {
        v.IdentNr: as_phx_obj(v, "PhxPolygon", _vertix_dict=vertix_dict) for v in geom_data.Polygons
    }

    # ------------------------------------------------------------------
    # -- Build the Components
    opaque_compo_dicts, ap_compo_dicts = _get_compo_data_by_type(bldg_data.Components)

    # -- Build all of the Aperture Components first
    aperture_dict: Dict[Tuple[int, ...], PhxComponentAperture] = {}
    for component_dict in ap_compo_dicts:
        ap_component: PhxComponentAperture = as_phx_obj(
            component_dict,
            "PhxComponentAperture",
            _polygons=polygon_dict,
            _window_types=_phx_project_host.window_types,
        )
        aperture_dict[ap_component.polygon_ids_sorted] = ap_component

    # -- Build all of the Opaque Components
    for component_dict in opaque_compo_dicts:
        op_component: PhxComponentOpaque = as_phx_obj(
            component_dict,
            "PhxComponentOpaque",
            _polygons=polygon_dict,
            _assembly_types=_phx_project_host.assembly_types,
            _aperture_dict=aperture_dict,
        )
        phx_obj.add_component(op_component)

    # ------------------------------------------------------------------
    # -- Build and add the Zones
    for zone_data_dict in bldg_data.Zones:
        new_zone = as_phx_obj(zone_data_dict, "PhxZone", _phx_project_host=_phx_project_host)
        phx_obj.add_zone(new_zone)

    return phx_obj


def _PhxVertix(_data: wufi_xml.WufiVertix) -> PhxVertix:
    phx_obj = PhxVertix()
    phx_obj.id_num = _data.IdentNr
    phx_obj.x = _data.X
    phx_obj.y = _data.Y
    phx_obj.z = _data.Z
    return phx_obj


def _PhxPolygon(_data: wufi_xml.WufiPolygon, _vertix_dict: Dict[int, PhxVertix] = {}) -> PhxPolygon:
    surface_normal = PhxVector(
        float(_data.NormalVectorX),
        float(_data.NormalVectorY),
        float(_data.NormalVectorZ),
    )

    # ----------------------------------------------------------------------
    plane = PhxPlane(
        normal_vector=surface_normal,
        origin=PhxVertix(0, 0, 0),  # temp. value
        x=PhxVector(0, 0, 0),  # temp. value
        y=PhxVector(0, 0, 0),  # temp. value
    )
    # ----------------------------------------------------------------------

    phx_obj = PhxPolygon(
        _display_name=str(_data.IdentNr),
        _area=None,
        _center=None,
        normal_vector=plane.normal_vector,
        plane=plane,
    )
    phx_obj.id_num = _data.IdentNr
    phx_obj._vertices = [_vertix_dict[_.IdentNr] for _ in _data.IdentNrPoints]
    if _data.IdentNrPolygonsInside:
        for _ in _data.IdentNrPolygonsInside:
            phx_obj.child_polygon_ids.append(_.IdentNr)

    # -- Figure out the Polygon Plane's x, y, and origin values based on the vertices
    phx_obj.plane.origin = phx_obj.vertices[0]
    phx_obj.plane.x = PhxVector.from_2_points(phx_obj.vertices[0], phx_obj.vertices[1])
    phx_obj.plane.y = phx_obj.plane.x.rotate_around(phx_obj.plane.normal_vector, 90)

    # -- Make sure to do this, otherwise area won't work properly.
    phx_obj.plane.x.unitize()
    phx_obj.plane.y.unitize()

    return phx_obj


def _PhxComponentAperture(
    _data: wufi_xml.WufiComponent,
    _polygons: Dict[int, PhxPolygon],
    _window_types: Dict[str, PhxConstructionWindow],
) -> PhxComponentAperture:
    phx_obj = PhxComponentAperture(_host=None)  # type: ignore

    phx_obj.display_name = _data.Name or "__unnamed_component__"
    phx_obj.face_type = ComponentFaceType.WINDOW
    phx_obj.face_opacity = ComponentFaceOpacity(_data.Type)
    phx_obj.color_interior = ComponentColor(_data.IdentNrColorI)
    phx_obj.color_exterior = ComponentColor(_data.IdentNrColorE)
    phx_obj.exposure_interior = _data.InnerAttachment
    phx_obj.exposure_exterior = ComponentExposureExterior(_data.OuterAttachment)

    if _data.IdentNr_ComponentInnerSurface:
        phx_obj.interior_attachment_id = _data.IdentNr_ComponentInnerSurface

    if _data.IdentNrWindowType:
        phx_obj.window_type = _window_types[str(_data.IdentNrWindowType)]
        phx_obj.window_type.id_num_shade = _data.IdentNrSolarProtection

    phx_obj.install_depth = _data.DepthWindowReveal
    phx_obj.default_monthly_shading_correction_factor = _data.DefaultCorrectionShadingMonth

    for i, poly_id in enumerate(_data.IdentNrPolygons):
        polygon = _polygons[poly_id.IdentNr]
        new_ap_element = PhxApertureElement(_host=phx_obj)
        new_ap_element.display_name = f"{phx_obj.display_name} - {i}"
        new_ap_element.polygon = polygon  # TODO: make a to_rect method for polygons....
        phx_obj.add_element(new_ap_element)

    return phx_obj


def _PhxComponentOpaque(
    _data: wufi_xml.WufiComponent,
    _polygons: Dict[int, PhxPolygon],
    _assembly_types: Dict[str, PhxConstructionOpaque],
    _aperture_dict: Dict[Tuple[int, ...], PhxComponentAperture],
) -> PhxComponentOpaque:
    # -- Build either an opaque or transparent component
    phx_obj = PhxComponentOpaque()

    phx_obj._id_num = _data.IdentNr
    phx_obj.display_name = _data.Name or "__unnamed_component__"
    phx_obj.face_opacity = ComponentFaceOpacity(_data.Type)
    phx_obj.color_interior = ComponentColor(_data.IdentNrColorI)
    phx_obj.color_exterior = ComponentColor(_data.IdentNrColorE)
    phx_obj.exposure_interior = _data.InnerAttachment
    phx_obj.exposure_exterior = ComponentExposureExterior(_data.OuterAttachment)

    if _data.IdentNr_ComponentInnerSurface:
        phx_obj.interior_attachment_id = _data.IdentNr_ComponentInnerSurface

    phx_obj.add_polygons([_polygons[i.IdentNr] for i in _data.IdentNrPolygons])
    if phx_obj.polygons:
        face_normal_angle_off_vertical = phx_obj.polygons[0].angle_from_horizontal
        phx_obj.face_type = ComponentFaceType.by_angle(face_normal_angle_off_vertical)

    # -- Add the Apertures
    # -- For each polygon, see if it has any children polygons? If so, find the
    # -- aperture element associated with the child polygon, and add it to the component.
    for poly in phx_obj.polygons:
        for child_id in poly.child_polygon_ids:
            for k, v in _aperture_dict.items():
                if child_id in k:
                    phx_obj.add_aperture(v)

    if _data.IdentNrAssembly:
        phx_obj.assembly_type_id_num = _data.IdentNrAssembly
        if phx_obj.assembly_type_id_num != -1:
            # -- -1 indicates a shade component without an assembly type
            phx_obj.assembly = _assembly_types[str(_data.IdentNrAssembly)]

    return phx_obj


def _PhxComponentThermalBridge(
    _data: wufi_xml.WufiThermalBridge,
) -> PhxComponentThermalBridge:
    phx_obj = PhxComponentThermalBridge()

    phx_obj.quantity = 1.0
    phx_obj.display_name = _data.Name
    phx_obj.group_type = ThermalBridgeType(_data.Type * -1)
    phx_obj.length = _data.Length
    phx_obj.psi_value = _data.PsiValue

    return phx_obj


# -----------------------------------------------------------------------------
# -- Foundations


def _PhxHeatedBasement(_data: wufi_xml.WufiFoundationInterface) -> PhxHeatedBasement:
    phx_obj = PhxHeatedBasement()
    phx_obj.display_name = _data.Name or ""
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.floor_slab_area_m2 = _data.FloorSlabArea
    phx_obj.floor_slab_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.floor_slab_u_value = _data.U_ValueBasementSlab
    phx_obj.slab_depth_below_grade_m = _data.DepthBasementBelowGroundSurface
    phx_obj.basement_wall_u_value = _data.U_ValueBasementWall
    return phx_obj


def _PhxUnHeatedBasement(_data: wufi_xml.WufiFoundationInterface) -> PhxUnHeatedBasement:
    phx_obj = PhxUnHeatedBasement()
    phx_obj.display_name = _data.Name or ""
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.slab_depth_below_grade_m = _data.DepthBasementBelowGroundSurface
    phx_obj.basement_wall_height_above_grade_m = _data.HeightBasementWallAboveGrade
    phx_obj.floor_ceiling_area_m2 = _data.FloorSlabArea
    phx_obj.floor_slab_u_value = _data.U_ValueBasementSlab
    phx_obj.floor_ceiling_area_m2 = _data.FloorCeilingArea
    phx_obj.ceiling_u_value = _data.U_ValueCeilingToUnheatedCellar
    phx_obj.basement_wall_uValue_below_grade = _data.U_ValueBasementWall
    phx_obj.basement_wall_uValue_above_grade = _data.U_ValueWallAboveGround
    phx_obj.floor_slab_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.basement_volume_m3 = _data.BasementVolume
    return phx_obj


def _PhxSlabOnGrade(_data: wufi_xml.WufiFoundationInterface) -> PhxSlabOnGrade:
    phx_obj = PhxSlabOnGrade()
    phx_obj.display_name = _data.Name or ""
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.floor_slab_area_m2 = _data.FloorSlabArea
    phx_obj.floor_slab_u_value = _data.U_ValueBasementSlab
    phx_obj.floor_slab_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.perim_insulation_position = _data.PositionPerimeterInsulation
    phx_obj.perim_insulation_width_or_depth_m = _data.PerimeterInsulationWidthDepth
    phx_obj.perim_insulation_thickness_m = _data.ThicknessPerimeterInsulation
    phx_obj.perim_insulation_conductivity = _data.ConductivityPerimeterInsulation
    return phx_obj


def _PhxVentedCrawlspace(_data: wufi_xml.WufiFoundationInterface) -> PhxVentedCrawlspace:
    phx_obj = PhxVentedCrawlspace()
    phx_obj.display_name = _data.Name or ""
    phx_obj.foundation_type_num = _data.FloorSlabType

    phx_obj.crawlspace_floor_slab_area_m2 = _data.FloorCeilingArea
    phx_obj.ceiling_above_crawlspace_u_value = _data.U_ValueCeilingToUnheatedCellar
    phx_obj.crawlspace_floor_exposed_perimeter_m = _data.FloorSlabPerimeter
    phx_obj.crawlspace_wall_height_above_grade_m = _data.HeightBasementWallAboveGrade
    phx_obj.crawlspace_floor_u_value = _data.U_ValueCrawlspaceFloor
    phx_obj.crawlspace_vent_opening_are_m2 = _data.CrawlspaceVentOpenings
    phx_obj.crawlspace_wall_u_value = _data.U_ValueWallAboveGround
    return phx_obj


def _PhxFoundation(_data: wufi_xml.WufiFoundationInterface) -> Optional[PhxFoundation]:
    foundation_type_builders = {
        FoundationType.HEATED_BASEMENT: "PhxHeatedBasement",
        FoundationType.UNHEATED_BASEMENT: "PhxUnHeatedBasement",
        FoundationType.SLAB_ON_GRADE: "PhxSlabOnGrade",
        FoundationType.VENTED_CRAWLSPACE: "PhxVentedCrawlspace",
        FoundationType.NONE: None,
    }

    device_type = FoundationType(_data.FloorSlabType)
    builder_class = foundation_type_builders[device_type]
    if not builder_class:
        return None

    # -- Pass the data off to the correct foundation builder class
    return as_phx_obj(_data, builder_class)


# -----------------------------------------------------------------------------
# -- Site & Climate


def _PhxPhiusCertification(_data: wufi_xml.WufiPassivehouseData) -> PhxPhiusCertification:
    phx_obj = PhxPhiusCertification()
    phx_obj.use_monthly_shading = _data.UseWUFIMeanMonthShading

    # no idea why PH_BuildingData is a list?....
    bldg_data = _data.PH_Buildings[0]
    phx_obj.ph_building_data = as_phx_obj(bldg_data, "PhxPhBuildingData")

    # ----------------------------------------------------------------------
    # --- Certification Criteria
    criteria = phx_obj.phius_certification_criteria
    criteria.ph_selection_target_data = _data.PH_SelectionTargetData
    criteria.phius_annual_heating_demand = _data.AnnualHeatingDemand
    criteria.phius_annual_cooling_demand = _data.AnnualCoolingDemand
    criteria.phius_peak_heating_load = _data.PeakHeatingLoad
    criteria.phius_peak_cooling_load = _data.PeakCoolingLoad

    # ----------------------------------------------------------------------
    # --- Certification Settings
    settings = phx_obj.phius_certification_settings
    settings.phius_building_certification_program = PhiusCertificationProgram(_data.PH_CertificateCriteria)
    settings.phius_building_category_type = PhiusCertificationBuildingCategoryType(bldg_data.BuildingCategory)
    if bldg_data.OccupancyTypeResidential:
        settings.phius_building_use_type = PhiusCertificationBuildingUseType(bldg_data.OccupancyTypeResidential)
    settings.phius_building_status = PhiusCertificationBuildingStatus(bldg_data.BuildingStatus)
    settings.phius_building_type = PhiusCertificationBuildingType(bldg_data.BuildingType)

    return phx_obj


def _PhxPhBuildingData(_data: wufi_xml.WufiPH_Building) -> PhxPhBuildingData:
    phx_obj = PhxPhBuildingData()
    phx_obj.id_num = _data.IdentNr
    phx_obj.num_of_units = _data.NumberUnits
    phx_obj.num_of_floors = _data.CountStories
    phx_obj.occupancy_setting_method = _data.OccupancySettingMethod
    phx_obj.airtightness_q50 = _data.EnvelopeAirtightnessCoefficient
    phx_obj.setpoints.winter = _data.IndoorTemperature
    phx_obj.setpoints.summer = _data.OverheatingTemperatureThreshold
    phx_obj.mech_room_temp = _data.MechanicalRoomTemperature
    phx_obj.non_combustible_materials = _data.NonCombustibleMaterials
    phx_obj.building_exposure_type = WindExposureType(_data.BuildingWindExposure)
    phx_obj.summer_hrv_bypass_mode = hvac_enums.PhxSummerBypassMode(_data.SummerHRVHumidityRecovery)

    for foundation_data in _data.FoundationInterfaces or []:
        phx_obj.add_foundation(as_phx_obj(foundation_data, "PhxFoundation"))

    return phx_obj


def _PhxSite(_data: wufi_xml.WufiClimateLocation) -> PhxSite:
    phx_obj = PhxSite()
    phx_obj.selection = SiteSelection(_data.Selection)

    ph_climate_data = _data.PH_ClimateLocation
    if ph_climate_data:
        phx_obj.location.latitude = ph_climate_data.Latitude
        phx_obj.location.longitude = ph_climate_data.Longitude
        phx_obj.location.hours_from_UTC = ph_climate_data.dUTC
        phx_obj.location.site_elevation = ph_climate_data.HeightNNBuilding
        phx_obj.location.climate_zone = ph_climate_data.ClimateZone

    phx_obj.climate = as_phx_obj(ph_climate_data, "PhxClimate")
    phx_obj.ground = as_phx_obj(ph_climate_data, "PhxGround")
    phx_obj.energy_factors = as_phx_obj(ph_climate_data, "PhxSiteEnergyFactors")

    return phx_obj


def _PhxClimate(_data: wufi_xml.WufiPH_ClimateLocation) -> PhxClimate:
    def _monthly_values(_in) -> List[float]:
        values = []
        for i in range(12):
            try:
                val = round(getattr(_in[i], "Item", 0.0), 2)
            except IndexError:
                val = 0.0
            values.append(val)
        return values

    phx_obj = PhxClimate()

    phx_obj.station_elevation = _data.HeightNNWeatherStation
    phx_obj.selection = SiteClimateSelection(_data.Selection)
    phx_obj.daily_temp_swing = _data.DailyTemperatureSwingSummer
    phx_obj.avg_wind_speed = _data.AverageWindSpeed

    phx_obj.temperature_air = _monthly_values(_data.TemperatureMonthly)
    phx_obj.temperature_dewpoint = _monthly_values(_data.DewPointTemperatureMonthly)
    phx_obj.temperature_sky = _monthly_values(_data.SkyTemperatureMonthly)
    phx_obj.radiation_north = _monthly_values(_data.NorthSolarRadiationMonthly)
    phx_obj.radiation_east = _monthly_values(_data.EastSolarRadiationMonthly)
    phx_obj.radiation_south = _monthly_values(_data.SouthSolarRadiationMonthly)
    phx_obj.radiation_west = _monthly_values(_data.WestSolarRadiationMonthly)
    phx_obj.radiation_global = _monthly_values(_data.GlobalSolarRadiationMonthly)

    phx_obj.peak_heating_1.temperature_air = _data.TemperatureHeating1
    phx_obj.peak_heating_1.radiation_north = _data.NorthSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_east = _data.EastSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_south = _data.SouthSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_west = _data.WestSolarRadiationHeating1
    phx_obj.peak_heating_1.radiation_global = _data.GlobalSolarRadiationHeating1

    phx_obj.peak_heating_2.temperature_air = _data.TemperatureHeating2
    phx_obj.peak_heating_2.radiation_north = _data.NorthSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_east = _data.EastSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_south = _data.SouthSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_west = _data.WestSolarRadiationHeating2
    phx_obj.peak_heating_2.radiation_global = _data.GlobalSolarRadiationHeating2

    phx_obj.peak_cooling_1.temperature_air = _data.TemperatureCooling
    phx_obj.peak_cooling_1.radiation_north = _data.NorthSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_east = _data.EastSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_south = _data.SouthSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_west = _data.WestSolarRadiationCooling
    phx_obj.peak_cooling_1.radiation_global = _data.GlobalSolarRadiationCooling

    phx_obj.peak_cooling_2.temperature_air = _data.TemperatureCooling2
    phx_obj.peak_cooling_2.radiation_north = _data.NorthSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_east = _data.EastSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_south = _data.SouthSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_west = _data.WestSolarRadiationCooling2
    phx_obj.peak_cooling_2.radiation_global = _data.GlobalSolarRadiationCooling2

    return phx_obj


def _PhxGround(_data: wufi_xml.WufiPH_ClimateLocation) -> PhxGround:
    phx_obj = PhxGround()
    phx_obj.ground_thermal_conductivity = _data.GroundThermalConductivity
    phx_obj.ground_heat_capacity = _data.GroundHeatCapacitiy
    phx_obj.ground_density = _data.GroundDensity
    phx_obj.depth_groundwater = _data.DepthGroundwater
    phx_obj.flow_rate_groundwater = _data.FlowRateGroundwater
    return phx_obj


def _PhxSiteEnergyFactors(_data: wufi_xml.WufiPH_ClimateLocation) -> PhxSiteEnergyFactors:
    _wufi_order = (
        "OIL",
        "NATURAL_GAS",
        "LPG",
        "HARD_COAL",
        "WOOD",
        "ELECTRICITY_MIX",
        "ELECTRICITY_PV",
        "HARD_COAL_CGS_70_CHP",
        "HARD_COAL_CGS_35_CHP",
        "HARD_COAL_CGS_0_CHP",
        "GAS_CGS_70_CHP",
        "GAS_CGS_35_CHP",
        "GAS_CGS_0_CHP",
        "OIL_CGS_70_CHP",
        "OIL_CGS_35_CHP",
        "OIL_CGS_0_CHP",
    )
    phx_obj = PhxSiteEnergyFactors()
    phx_obj.selection_pe_co2_factor = SiteEnergyFactorSelection(_data.SelectionPECO2Factor)

    # If any of the 'Standard' set are used, their data will NOT be included in
    # in the WUFI XML file. So set them based on the Phius Cert Program being used.
    #
    # If User-defined factors are used, they will be included in the XML file.
    # So in that case, use those values for the PE and CO2 factors.
    for name, factor in zip(_wufi_order, _data.PEFactorsUserDef or []):
        phx_obj.pe_factors[name] = PhxPEFactor(factor.__root__, "kWh/kWh", name)

    for name, factor in zip(_wufi_order, _data.CO2FactorsUserDef or []):
        phx_obj.co2_factors[name] = PhxCO2Factor(factor.__root__, "g/kWh", name)

    return phx_obj


# -----------------------------------------------------------------------------
# -- Zones


def _PhxZone(_data: wufi_xml.WufiZone, _phx_project_host: PhxProject) -> PhxZone:
    def _spec_cap_WH_m2k(_input):
        if _input <= 60.0:
            return 1
        elif _input <= 132.0:
            return 2
        else:
            return 3

    phx_obj = PhxZone()

    phx_obj.zone_type = ZoneType(_data.KindZone)
    phx_obj.attached_zone_type = AttachedZoneType(_data.KindAttachedZone or 0)
    phx_obj.attached_zone_reduction_factor = _data.TemperatureReductionFactorUserDefined or 1.0
    phx_obj.id_num = _data.IdentNr
    phx_obj.display_name = _data.Name
    phx_obj.volume_gross = _data.GrossVolume or 0.0
    phx_obj.volume_net = _data.NetVolume or 0.0
    phx_obj.weighted_net_floor_area = _data.FloorArea or 0.0
    phx_obj.clearance_height = _data.ClearanceHeight or 0.0
    phx_obj.res_occupant_quantity = _data.OccupantQuantityUserDef
    phx_obj.res_number_bedrooms = _data.NumberBedrooms or 0
    phx_obj.specific_heat_capacity = SpecificHeatCapacity(_spec_cap_WH_m2k(_data.SpecificHeatCapacity))

    # ----------------------------------------------------------------------
    # -- Create all the spaces from the XML "RoomsVentilation" data
    for space_ventilation_data in _data.RoomsVentilation or []:
        phx_obj.spaces.append(
            as_phx_obj(
                space_ventilation_data,
                "PhxSpace",
                _phx_project_host=_phx_project_host,
            )
        )

    # -- Try and add in any occupancy, lighting loads to the spaces, if they exist.
    # -- Will try and find a load with a matching name.
    occupancy_load_data = {d.Name: d for d in _data.LoadsPersonsPH or []}
    lighting_load_data = {d.Name: d for d in _data.LoadsLightingsPH or []}
    for space in phx_obj.spaces:
        space = _add_occupancy_data_to_space(
            space, _phx_project_host, occupancy_load_data.get(space.display_name, None)
        )
        space = _add_lighting_data_to_space(space, _phx_project_host, lighting_load_data.get(space.display_name, None))

    # ----------------------------------------------------------------------
    # -- Add in any devices and thermal bridges as well.
    for vent_dict_data in _data.ExhaustVents or []:
        new_phx_exhaust = as_phx_obj(vent_dict_data, "PhxExhaustVent")
        phx_obj.exhaust_ventilator_collection.add_new_ventilator(new_phx_exhaust.identifier, new_phx_exhaust)

    for device_dict_data in _data.HomeDevice or []:
        new_phx_home_device = as_phx_obj(device_dict_data, "PhxHomeDevice")
        phx_obj.elec_equipment_collection.add_new_device(new_phx_home_device.identifier, new_phx_home_device)

    for i, bridge_dict_data in enumerate(_data.ThermalBridges or []):
        new_phx_tb = as_phx_obj(bridge_dict_data, "PhxComponentThermalBridge")
        new_phx_tb.identifier = f"ThermalBridge_{i :03}"
        phx_obj.add_thermal_bridge(new_phx_tb)

    return phx_obj


def _add_occupancy_data_to_space(
    _space: PhxSpace,
    _phx_project_host: PhxProject,
    _occupancy_data: Optional[wufi_xml.WufiLoadPerson],
) -> PhxSpace:
    """Add occupancy data to a space, if any is supplied."""
    if not _occupancy_data:
        return _space

    # -- Get the Project's Occupancy Utilization Pattern
    project_util_patterns_occ = _phx_project_host.utilization_patterns_occupancy

    # -- Ensure that the Space's Occupancy pattern is in the project collection. If
    # -- not, create a new one.
    occ_pattern_id = str(_occupancy_data.IdentNrUtilizationPattern)
    if not project_util_patterns_occ.key_is_in_collection(occ_pattern_id):
        new_schedule = PhxScheduleOccupancy.constant_operation()
        new_schedule.display_name = _occupancy_data.Name
        new_schedule.id_num = int(occ_pattern_id)
        new_schedule.identifier = str(occ_pattern_id)
        project_util_patterns_occ.add_new_util_pattern(new_schedule)

    # -- Assign the occupancy pattern to the space.
    _space.occupancy.schedule = project_util_patterns_occ[str(occ_pattern_id)]

    # -- Set the occupancy load for the space.
    _space.peak_occupancy = _occupancy_data.NumberOccupants

    return _space


def _add_lighting_data_to_space(
    _space: PhxSpace,
    _phx_project_host: PhxProject,
    _lighting_data: Optional[wufi_xml.WufiLoadsLighting],
) -> PhxSpace:
    """Add lighting data to a space, if any is supplied."""
    if not _lighting_data:
        return _space

    # -- Get the right Lighting Utilization Pattern
    project_util_pat_lighting = _phx_project_host.utilization_patterns_lighting
    lighting_pattern_id = _lighting_data.RoomCategory

    # -- Ensure that the lighting pattern is in the project collection. If
    # -- not, create a new one based on the full-load lighting hours.
    if not project_util_pat_lighting.key_is_in_collection(lighting_pattern_id):
        new_schedule = PhxScheduleLighting.from_annual_operating_hours(_lighting_data.LightingFullLoadHours)
        new_schedule.display_name = _lighting_data.Name
        new_schedule.id_num = int(lighting_pattern_id)
        new_schedule.identifier = str(lighting_pattern_id)
        project_util_pat_lighting.add_new_util_pattern(new_schedule)

    # -- Assign the lighting pattern to the space.
    _space.lighting.schedule = project_util_pat_lighting[str(lighting_pattern_id)]

    # -- Set the lighting load for the space.
    _space.lighting.load.installed_w_per_m2 = _lighting_data.InstalledLightingPower

    return _space


def _PhxSpace(_data: wufi_xml.WufiRoom, _phx_project_host: PhxProject) -> PhxSpace:
    phx_obj = PhxSpace()

    # phx_obj.id_num = int(_data."])
    phx_obj.display_name = _data.Name
    phx_obj.wufi_type = _data.Type
    phx_obj.quantity = _data.Quantity
    phx_obj.floor_area = _data.AreaRoom or 0.0
    phx_obj.weighted_floor_area = _data.AreaRoom or 0.0
    phx_obj.clear_height = _data.ClearRoomHeight or 0.0

    # -- Ventilation Unit (ERV) number
    phx_obj.vent_unit_id_num = _data.IdentNrVentilationUnit

    # -- Ventilation Schedule and Load
    project_util_pat_ven = _phx_project_host.utilization_patterns_ventilation
    vent_pattern_id = _data.IdentNrUtilizationPatternVent
    phx_obj.ventilation.schedule = project_util_pat_ven[str(vent_pattern_id)]
    phx_obj.ventilation.load.flow_extract = _data.DesignVolumeFlowRateExhaust
    phx_obj.ventilation.load.flow_supply = _data.DesignVolumeFlowRateSupply

    return phx_obj


# -----------------------------------------------------------------------------
# -- Mechanical Systems and Devices


def _PhxZoneCoverage(_data: wufi_xml.WufiZoneCoverage) -> PhxZoneCoverage:
    phx_obj = PhxZoneCoverage()

    phx_obj.zone_num = _data.IdentNrZone
    phx_obj.heating = _data.CoverageHeating
    phx_obj.cooling = _data.CoverageCooling
    phx_obj.ventilation = _data.CoverageVentilation
    phx_obj.humidification = _data.CoverageHumidification
    phx_obj.dehumidification = _data.CoverageDehumidification

    return phx_obj


def _PhxMechanicalDevice(_data: wufi_xml.WufiDevice) -> Any:
    device_builder_class_names = {
        hvac_enums.SystemType.VENTILATION: "PhxDevice_Ventilation",
        hvac_enums.SystemType.ELECTRIC: "PhxDevice_Electric",
        hvac_enums.SystemType.BOILER: "PhxDevice_Boiler",
        hvac_enums.SystemType.DISTRICT_HEAT: "PhxDevice_DistrictHeat",
        hvac_enums.SystemType.HEAT_PUMP: "PhxDevice_HeatPump",
        hvac_enums.SystemType.WATER_STORAGE: "PhxDevice_WaterStorage",
        hvac_enums.SystemType.PHOTOVOLTAIC: "PhxDevice_Photovoltaic",
    }

    # -- Find the right device-builder-classname
    try:
        system_type = hvac_enums.SystemType(_data.SystemType)
    except:
        print(f"Error: I do not understand the mech device type: '{_data.SystemType}'?")
        return None
    builder_class_name = device_builder_class_names[system_type]

    # -- Pass the data off using the correct device-builder-classname
    new_mech_device = as_phx_obj(_data, builder_class_name)

    # -- Set the usage profile (heating %, cooling %, etc.) of the device
    new_mech_device.usage_profile = as_phx_obj(_data, "PhxUsageProfile")

    return new_mech_device


def _PhxDevice_Ventilation(_data: wufi_xml.WufiDevice) -> PhxDeviceVentilator:
    phx_obj = PhxDeviceVentilator()

    phx_obj.display_name = _data.Name or "unnamed_ventilation"
    phx_obj.id_num = _data.IdentNr
    phx_obj.identifier = str(_data.IdentNr)

    phx_obj.params.sensible_heat_recovery = _data.HeatRecovery
    phx_obj.params.latent_heat_recovery = _data.MoistureRecovery

    if _data.PH_Parameters:
        phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace or False
        phx_obj.params.quantity = _data.PH_Parameters.Quantity or 0
        phx_obj.params.electric_efficiency = _data.PH_Parameters.ElectricEfficiency or 0.0
        phx_obj.params.frost_protection_reqd = _data.PH_Parameters.FrostProtection
        phx_obj.params.temperature_below_defrost_used = _data.PH_Parameters.TemperatureBelowDefrostUsed or 0.0

    return phx_obj


def _PhxDevice_Electric(_data: wufi_xml.WufiDevice) -> PhxHeaterElectric:
    phx_obj = PhxHeaterElectric()

    phx_obj.display_name = _data.Name or "unnamed_direct_electric"
    return phx_obj


def _PhxDevice_Boiler(_data: wufi_xml.WufiDevice) -> Optional[AnyPhxHeaterBoiler]:
    boiler_builders = {
        hvac_enums.PhxFuelType.NATURAL_GAS: "PhxHeaterBoilerFossil",
        hvac_enums.PhxFuelType.OIL: "PhxHeaterBoilerFossil",
        hvac_enums.PhxFuelType.WOOD_LOG: "PhxHeaterBoilerWood",
        hvac_enums.PhxFuelType.WOOD_PELLET: "PhxHeaterBoilerWood",
    }
    if not _data.PH_Parameters:
        return None
    boiler_type = hvac_enums.PhxFuelType(_data.PH_Parameters.EnergySourceBoilerType)
    return as_phx_obj(_data, boiler_builders[boiler_type])


def _PhxHeaterBoilerFossil(_data: wufi_xml.WufiDevice) -> PhxHeaterBoilerFossil:
    phx_obj = PhxHeaterBoilerFossil()

    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_fossil_fuel_boiler"
    phx_obj.params.fuel = _data.PH_Parameters.EnergySourceBoilerType
    phx_obj.params.condensing = _data.PH_Parameters.CondensingBoiler
    phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace
    phx_obj.params.effic_at_30_percent_load = _data.PH_Parameters.BoilerEfficiency30
    phx_obj.params.effic_at_nominal_load = _data.PH_Parameters.BoilerEfficiencyNominalOutput
    phx_obj.params.avg_rtrn_temp_at_30_percent_load = _data.PH_Parameters.AverageReturnTemperatureMeasured30Load
    phx_obj.params.avg_temp_at_70C_55C = _data.PH_Parameters.AverageBoilerTemperatureDesign70_55
    phx_obj.params.avg_temp_at_55C_45C = _data.PH_Parameters.AverageBoilerTemperatureDesign55_45
    phx_obj.params.avg_temp_at_32C_28C = _data.PH_Parameters.AverageBoilerTemperatureDesign35_28
    phx_obj.params.standby_loss_at_70C = _data.PH_Parameters.StandbyHeatLossBoiler70
    phx_obj.params.rated_capacity = _data.PH_Parameters.MaximalBoilerPower
    return phx_obj


def _PhxHeaterBoilerWood(_data: wufi_xml.WufiDevice) -> PhxHeaterBoilerWood:
    phx_obj = PhxHeaterBoilerWood()

    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_wood_boiler"
    return phx_obj


def _PhxDevice_DistrictHeat(_data: wufi_xml.WufiDevice) -> PhxHeaterDistrictHeat:
    phx_obj = PhxHeaterDistrictHeat()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_district_heat"
    return phx_obj


def _PhxDevice_HeatPump(_data: wufi_xml.WufiDevice) -> Optional[PhxHeatPumpDevice]:
    hp_builders = {
        hvac_enums.HeatPumpType.COMBINED: "PhxDevice_HeatPump_Combined",
        hvac_enums.HeatPumpType.ANNUAL: "PhxDevice_HeatPump_Annual",
        hvac_enums.HeatPumpType.RATED_MONTHLY: "PhxDevice_HeatPump_RatedMonthly",
        hvac_enums.HeatPumpType.HOT_WATER: "PhxDevice_HeatPump_HotWater",
    }
    if _data.PH_Parameters and _data.PH_Parameters.HPType:
        hp_type = hvac_enums.HeatPumpType(_data.PH_Parameters.HPType)
        builder_class = hp_builders[hp_type]
        return as_phx_obj(_data, builder_class)
    else:
        return None


def _PhxDevice_HeatPump_Combined(_data: wufi_xml.WufiDevice) -> PhxHeatPumpCombined:
    phx_obj = PhxHeatPumpCombined()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_combined_heat_pump"
    return phx_obj


def _PhxDevice_HeatPump_Annual(_data: wufi_xml.WufiDevice) -> PhxHeatPumpAnnual:
    phx_obj = PhxHeatPumpAnnual()

    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_annual_heat_pump"
    phx_obj.params.annual_COP = _data.PH_Parameters.AnnualCOP
    phx_obj.params.total_system_perf_ratio = _data.PH_Parameters.TotalSystemPerformanceRatioHeatGenerator
    return phx_obj


def _PhxDevice_HeatPump_RatedMonthly(_data: wufi_xml.WufiDevice) -> PhxHeatPumpMonthly:
    phx_obj = PhxHeatPumpMonthly()

    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_monthly_heat_pump"
    phx_obj.params.COP_1 = _data.PH_Parameters.RatedCOP1
    phx_obj.params.COP_2 = _data.PH_Parameters.RatedCOP2
    phx_obj.params.ambient_temp_1 = _data.PH_Parameters.AmbientTemperature1
    phx_obj.params.ambient_temp_2 = _data.PH_Parameters.AmbientTemperature2

    return phx_obj


def _PhxDevice_HeatPump_HotWater(_data: wufi_xml.WufiDevice) -> PhxHeatPumpHotWater:
    phx_obj = PhxHeatPumpHotWater()

    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_hot_water_heat_pump"
    phx_obj.params.annual_COP = _data.PH_Parameters.AnnualCOP
    phx_obj.params.total_system_perf_ratio = _data.PH_Parameters.TotalSystemPerformanceRatioHeatGenerator
    phx_obj.params.annual_energy_factor = _data.PH_Parameters.HPWH_EF
    return phx_obj


def _PhxDevice_WaterStorage(_data: wufi_xml.WufiDevice) -> PhxHotWaterTank:
    phx_obj = PhxHotWaterTank()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_hot_water_tank"
    phx_obj.quantity = _data.PH_Parameters.QauntityWS
    phx_obj.params.display_name = _data.Name or "unnamed_hot_water_tank"
    phx_obj.params.input_option = hvac_enums.PhxHotWaterInputOptions(_data.PH_Parameters.InputOption)
    phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace
    phx_obj.params.storage_loss_rate = _data.PH_Parameters.AverageHeatReleaseStorage

    phx_obj.params.storage_capacity = _data.PH_Parameters.SolarThermalStorageCapacity
    phx_obj.params.standby_losses = _data.PH_Parameters.StorageLossesStandby
    phx_obj.params.room_temp = _data.PH_Parameters.TankRoomTemp
    phx_obj.params.water_temp = _data.PH_Parameters.TypicalStorageWaterTemperature

    return phx_obj


def _PhxDevice_Photovoltaic(_data: wufi_xml.WufiDevice) -> PhxDevicePhotovoltaic:
    phx_obj = PhxDevicePhotovoltaic()
    if not _data.PH_Parameters:
        return phx_obj

    phx_obj.display_name = _data.Name or "unnamed_photovoltaic"
    phx_obj.params.location_type = _data.PH_Parameters.SelectionLocation
    phx_obj.params.onsite_utilization_type = _data.PH_Parameters.SelectionOnSiteUtilization

    phx_obj.params.utilization_type = _data.PH_Parameters.SelectionUtilization
    phx_obj.params.array_size = _data.PH_Parameters.ArraySizePV
    phx_obj.params.photovoltaic_renewable_energy = _data.PH_Parameters.PhotovoltaicRenewableEnergy

    phx_obj.params.onsite_utilization_factor = _data.PH_Parameters.OnsiteUtilization
    phx_obj.params.auxiliary_energy = _data.PH_Parameters.AuxiliaryEnergy
    phx_obj.params.auxiliary_energy_DHW = _data.PH_Parameters.AuxiliaryEnergyDHW
    phx_obj.params.in_conditioned_space = _data.PH_Parameters.InConditionedSpace

    return phx_obj


def _PhxUsageProfile(_data: wufi_xml.WufiDevice) -> PhxUsageProfile:
    new_phx_profile = PhxUsageProfile()

    if _data.Heating_Parameters:
        new_phx_profile.space_heating_percent = _data.Heating_Parameters.CoverageWithinSystem or 0.0

    if _data.Cooling_Parameters:
        new_phx_profile.cooling_percent = _data.Cooling_Parameters.CoverageWithinSystem or 0.0

    if _data.DHW_Parameters:
        new_phx_profile.dhw_heating_percent = _data.DHW_Parameters.CoverageWithinSystem or 0.0

    if _data.Ventilation_Parameters:
        new_phx_profile.ventilation_percent = _data.Ventilation_Parameters.CoverageWithinSystem or 0.0

    new_phx_profile.humidification_percent = _data.UsedFor_Humidification

    new_phx_profile.dehumidification_percent = _data.UsedFor_Dehumidification

    return new_phx_profile


# ----------------------------------------------------------------------
# -- Exhaust Ventilators


def _PhxExhaustVent(_data: wufi_xml.WufiExhaustVent) -> AnyPhxExhaustVent:
    exhaust_vent_builders = {
        hvac_enums.PhxExhaustVentType.KITCHEN_HOOD: "PhxExhaustVent_KitchenHood",
        hvac_enums.PhxExhaustVentType.DRYER: "PhxExhaustVent_Dryer",
        hvac_enums.PhxExhaustVentType.USER_DEFINED: "PhxExhaustVent_UserDefined",
    }
    vent_type = hvac_enums.PhxExhaustVentType(_data.Type)
    return as_phx_obj(_data, exhaust_vent_builders[vent_type])


def _PhxExhaustVent_KitchenHood(
    _data: wufi_xml.WufiExhaustVent,
) -> PhxExhaustVentilatorRangeHood:
    phx_obj = PhxExhaustVentilatorRangeHood()
    phx_obj.params.annual_runtime_minutes = _data.RunTimePerYear
    phx_obj.params.exhaust_flow_rate_m3h = _data.ExhaustVolumeFlowRate
    return phx_obj


def _PhxExhaustVent_Dryer(_data: wufi_xml.WufiExhaustVent) -> PhxExhaustVentilatorDryer:
    phx_obj = PhxExhaustVentilatorDryer()
    phx_obj.params.annual_runtime_minutes = _data.RunTimePerYear
    phx_obj.params.exhaust_flow_rate_m3h = _data.ExhaustVolumeFlowRate
    return phx_obj


def _PhxExhaustVent_UserDefined(
    _data: wufi_xml.WufiExhaustVent,
) -> PhxExhaustVentilatorUserDefined:
    phx_obj = PhxExhaustVentilatorUserDefined()
    phx_obj.params.annual_runtime_minutes = _data.RunTimePerYear
    phx_obj.params.exhaust_flow_rate_m3h = _data.ExhaustVolumeFlowRate
    return phx_obj


# ----------------------------------------------------------------------
# -- Electrical Devices


def _PhxHomeDevice(_data: wufi_xml.WufiHomeDevice) -> PhxElectricalDevice:
    device_builders = {
        ElectricEquipmentType.DISHWASHER: "PhxDeviceDishwasher",
        ElectricEquipmentType.CLOTHES_WASHER: "PhxDeviceClothesWasher",
        ElectricEquipmentType.CLOTHES_DRYER: "PhxDeviceClothesDryer",
        ElectricEquipmentType.REFRIGERATOR: "PhxDeviceRefrigerator",
        ElectricEquipmentType.FREEZER: "PhxDeviceFreezer",
        ElectricEquipmentType.FRIDGE_FREEZER: "PhxDeviceFridgeFreezer",
        ElectricEquipmentType.COOKING: "PhxDeviceCooktop",
        ElectricEquipmentType.CUSTOM: "PhxDeviceCustomElec",
        ElectricEquipmentType.MEL: "PhxDeviceMEL",
        ElectricEquipmentType.LIGHTING_INTERIOR: "PhxDeviceLightingInterior",
        ElectricEquipmentType.LIGHTING_EXTERIOR: "PhxDeviceLightingExterior",
        ElectricEquipmentType.LIGHTING_GARAGE: "PhxDeviceLightingGarage",
        ElectricEquipmentType.CUSTOM_LIGHTING: "PhxDeviceCustomLighting",
        ElectricEquipmentType.CUSTOM_MEL: "PhxDeviceCustomMEL",
    }
    device_type = ElectricEquipmentType(_data.Type)
    phx_obj: PhxElectricalDevice = as_phx_obj(_data, device_builders[device_type])

    # -- Set all the common parameters
    phx_obj.comment = _data.Comment
    phx_obj.reference_quantity = _data.ReferenceQuantity
    phx_obj.quantity = _data.Quantity
    phx_obj.in_conditioned_space = _data.InConditionedSpace
    phx_obj.reference_energy_norm = _data.ReferenceEnergyDemandNorm
    phx_obj.energy_demand = _data.EnergyDemandNorm
    phx_obj.energy_demand_per_use = _data.EnergyDemandNormUse
    phx_obj.combined_energy_factor = _data.CEF_CombinedEnergyFactor

    return phx_obj


def _PhxDeviceDishwasher(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceDishwasher:
    phx_obj = PhxDeviceDishwasher()
    phx_obj.water_connection = _data.Connection
    phx_obj.capacity_type = _data.DishwasherCapacityPreselection
    phx_obj.capacity = _data.DishwasherCapacityInPlace
    return phx_obj


def _PhxDeviceClothesWasher(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceClothesWasher:
    phx_obj = PhxDeviceClothesWasher()
    phx_obj.water_connection = _data.Connection
    phx_obj.utilization_factor = _data.UtilizationFactor
    phx_obj.capacity = _data.CapacityClothesWasher
    phx_obj.modified_energy_factor = _data.MEF_ModifiedEnergyFactor
    return phx_obj


def _PhxDeviceClothesDryer(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceClothesDryer:
    phx_obj = PhxDeviceClothesDryer()
    phx_obj.dryer_type = _data.Dryer_Choice or 4
    phx_obj.gas_consumption = _data.GasConsumption
    phx_obj.gas_efficiency_factor = _data.EfficiencyFactorGas
    phx_obj.field_utilization_factor_type = _data.FieldUtilizationFactorPreselection
    phx_obj.field_utilization_factor = _data.FieldUtilizationFactor
    return phx_obj


def _PhxDeviceRefrigerator(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceRefrigerator:
    phx_obj = PhxDeviceRefrigerator()
    return phx_obj


def _PhxDeviceFreezer(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceFreezer:
    phx_obj = PhxDeviceFreezer()
    return phx_obj


def _PhxDeviceFridgeFreezer(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceFridgeFreezer:
    phx_obj = PhxDeviceFridgeFreezer()
    return phx_obj


def _PhxDeviceCooktop(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceCooktop:
    phx_obj = PhxDeviceCooktop()
    phx_obj.cooktop_type = _data.CookingWith
    return phx_obj


def _PhxDeviceCustomElec(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceCustomElec:
    phx_obj = PhxDeviceCustomElec()
    return phx_obj


def _PhxDeviceMEL(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceMEL:
    phx_obj = PhxDeviceMEL()
    return phx_obj


def _PhxDeviceLightingInterior(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceLightingInterior:
    phx_obj = PhxDeviceLightingInterior()
    phx_obj.frac_high_efficiency = _data.FractionHightEfficiency
    return phx_obj


def _PhxDeviceLightingExterior(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceLightingExterior:
    phx_obj = PhxDeviceLightingExterior()
    phx_obj.frac_high_efficiency = _data.FractionHightEfficiency
    return phx_obj


def _PhxDeviceLightingGarage(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceLightingGarage:
    phx_obj = PhxDeviceLightingGarage()
    phx_obj.frac_high_efficiency = _data.FractionHightEfficiency
    return phx_obj


def _PhxDeviceCustomLighting(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceCustomLighting:
    phx_obj = PhxDeviceCustomLighting()

    return phx_obj


def _PhxDeviceCustomMEL(_data: wufi_xml.WufiHomeDevice) -> PhxDeviceCustomMEL:
    phx_obj = PhxDeviceCustomMEL()

    return phx_obj



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP Climate worksheet."""

from __future__ import annotations

from typing import List

from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import climate_entry
from PHX.xl import xl_app


class Climate:
    """IO Controller for the PHPP Climate Worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Climate):
        self.xl = _xl
        self.shape = _shape
        self.weather_data_start_rows: List[int] = []

    def get_start_rows(self) -> List[int]:
        # TODO: make this find the right starting rows.
        return [self.shape.ud_block.start_row]

    def write_climate_block(self, _climate_entry: climate_entry.ClimateDataBlock) -> None:
        if not self.weather_data_start_rows:
            self.weather_data_start_rows = self.get_start_rows()

        # Just use the first one for now....
        # TODO: Write all variants to different slots
        start_row = self.weather_data_start_rows[0]

        for item in _climate_entry.create_xl_items(self.shape.name, start_row):
            self.xl.write_xl_item(item)

    def write_active_climate(self, _active_climate: climate_entry.ClimateSettings) -> None:
        start_row = 9
        for item in _active_climate.create_xl_items(self.shape.name, start_row):
            self.xl.write_xl_item(item)

    def read_active_country(self) -> str:
        return str(self.xl.get_single_data_item(self.shape.name, self.shape.named_ranges.country))

    def read_active_region(self) -> str:
        return str(self.xl.get_single_data_item(self.shape.name, self.shape.named_ranges.region))

    def read_active_data_set(self) -> str:
        return str(self.xl.get_single_data_item(self.shape.name, self.shape.named_ranges.data_set))

    def read_station_elevation(self) -> str:
        return str(self.xl.get_single_data_item(self.shape.name, self.shape.defined_ranges.weather_station_altitude))

    def read_site_elevation(self) -> str:
        return str(self.xl.get_single_data_item(self.shape.name, self.shape.defined_ranges.site_altitude))

    def read_latitude(self) -> float:
        return float(self.xl.get_single_data_item(self.shape.name, self.shape.defined_ranges.latitude))

    def read_longitude(self) -> float:
        return float(self.xl.get_single_data_item(self.shape.name, self.shape.defined_ranges.longitude))



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Water Piping Distribution Objects."""

from __future__ import annotations

import math
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Union
from uuid import uuid4

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.polyline import LineSegment3D
from ph_units.converter import convert

from PHX.model.enums.hvac import (
    PhxHotWaterPipingCalcMethod,
    PhxHotWaterPipingInchDiameterType,
    PhxHotWaterPipingMaterial,
    PhxHotWaterSelectionUnitsOrFloors,
)


@dataclass
class PhxRecirculationParameters:
    calc_method: PhxHotWaterPipingCalcMethod = field(default=PhxHotWaterPipingCalcMethod.HOT_WATER_PIPING_FLOOR_METHOD)
    pipe_material: PhxHotWaterPipingMaterial = field(default=PhxHotWaterPipingMaterial.COPPER_L)
    demand_recirc: bool = True
    num_bathrooms: int = 1
    hot_water_fixtures: int = 1
    all_pipes_insulated: bool = True
    units_or_floors: PhxHotWaterSelectionUnitsOrFloors = field(
        default=PhxHotWaterSelectionUnitsOrFloors.USER_DETERMINED
    )
    pipe_diameter: float = 25.4  # MM - For simplified method only
    air_temp: float = 20.0  # Deg C
    water_temp: float = 60.0  # Deg C
    daily_recirc_hours: float = 0.0


@dataclass
class PhxPipeSegment:
    """An individual Pipe Segment."""

    identifier: str
    display_name: str
    geometry: LineSegment3D
    pipe_material: PhxHotWaterPipingMaterial
    diameter_m: float
    insulation_thickness_m: float
    insulation_conductivity: float  # W/mk
    insulation_reflective: bool
    insulation_quality: Any
    daily_period: float
    water_temp_c: float = 60.0
    pipe_wall_thickness_m: float = 0.00225

    @property
    def diameter_mm(self) -> float:
        return self.diameter_m * 1000

    @property
    def diameter_inner_m(self) -> float:
        return self.diameter_m

    @property
    def diameter_outer_m(self) -> float:
        """Return the outside diameter including the pipe wall thickness."""
        return self.diameter_inner_m + self.pipe_wall_thickness_m

    @property
    def diameter_with_insulation_m(self) -> float:
        return self.diameter_outer_m + (2 * self.insulation_thickness_m)

    @property
    def length_m(self) -> float:
        return self.geometry.length

    @property
    def pipe_heat_loss_coefficient(self) -> float:
        """Return the pipe's heat-loss-coefficient (W/mk) considering the diameter and insulation."""
        return self._solve_for_pipe_heat_loss_coeff()

    @property
    def _starting_alpha(self) -> float:
        """Return W/m2k based on insulation reflectivity."""
        if self.insulation_reflective:
            return 5.0
        else:
            return 8.0

    def _calc_pipe_heat_loss_coeff(self, _alpha: float, _conductivity: Optional[float] = None) -> float:
        """Calculate the pipe heat-loss coefficient (W/mk) with a known Alpha (W/m2k) value."""

        # -- Allow for 'reverse' solver....
        conductivity = _conductivity or self.insulation_conductivity

        _a = self.diameter_outer_m / self.diameter_inner_m
        _b = self.diameter_with_insulation_m / self.diameter_outer_m
        _c = (
            1 / 2 / 55 * math.log(_a)
            + 1 / 2 / conductivity * math.log(_b)
            + 1 / self.diameter_with_insulation_m / _alpha
        )
        return math.pi / _c

    def _calc_pipe_surface_temp(self, dT, _k, _conductivity: Optional[float] = None) -> float:
        """Return a surface temp (k) for the pipe with a known heat-loss-coefficient (W/mk) value."""

        # -- Allow for 'reverse' solver....
        conductivity = _conductivity or self.insulation_conductivity

        _a = self.diameter_outer_m / self.diameter_inner_m
        _b = self.diameter_with_insulation_m / self.diameter_outer_m
        _c = 1 / 2 / 55 * math.log(_a) + 1 / 2 / conductivity * math.log(_b)
        return dT - 1 / math.pi * _c * _k * dT

    def _solve_for_pipe_heat_loss_coeff(
        self,
        _alpha: Optional[float] = None,
        _k1: float = 100.0,
        _k2: float = 1.0,
        _surface_temp: Optional[float] = None,
    ) -> float:
        """Return a heat-loss coefficient (W/mk) considering the diameter and insulation.

        Approximates the recursive algorithm found in the PHPP "DHW" worksheet.
        """
        TOLERANCE = 0.001
        DELTA_T = 30  # K
        if self.insulation_reflective:
            N = 0.1
        else:
            N = 0.85

        while _k1 - _k2 > TOLERANCE:
            # -- hang onto the old k to test against tolerance later
            _k1 = _k2

            # -- If W/m2k alpha is not provided (ie: at the start) use the starting-alpha constant
            if not _alpha:
                _alpha = self._starting_alpha

            # -- Calc the new K (W/mk), and new surface temp of the pipe
            _k2 = self._calc_pipe_heat_loss_coeff(_alpha)
            _surface_temp = self._calc_pipe_surface_temp(DELTA_T, _k2)

            # -- calc new alpha W/m2k value
            _alpha = N * 4.8 + 1.62 * _surface_temp**0.333

            # -- Run the solver until the k result is < the tolerance
            self._solve_for_pipe_heat_loss_coeff(_alpha, _k1, _k2, _surface_temp)

        return _k2

    def reverse_solve_for_insulation_conductivity(
        self,
        target_result: float,
        starting_conductivity: float,
    ) -> float:
        """Return an insulation conductivity (W/mk) when given a known heat-loss-coeff (W/MK).

        Be sure to set the known (or assumed) pipe-diameter and insulation-thickness of the Segment before trying to solve.
        """
        TOLERANCE_A = 0.001
        TOLERANCE_B = 0.01
        DELTA_T = 30  # K

        if self.insulation_reflective:
            N = 0.1
        else:
            N = 0.85

        conductivity = starting_conductivity

        while True:
            alpha = 5.0  # Assuming a known starting value for _alpha
            k1 = 2.0  # Assuming a high initial value for _k1
            k2 = 0.01  # Assuming a low initial value for _k2
            surface_temp = None  # Placeholder for surface temp

            # -- Use the values and calc the heat-loss-coefficient...
            while k1 - k2 > TOLERANCE_A:
                k1 = k2
                k2 = self._calc_pipe_heat_loss_coeff(alpha, conductivity)
                surface_temp = self._calc_pipe_surface_temp(DELTA_T, k2, conductivity)
                alpha = N * 4.8 + 1.62 * surface_temp**0.333

            if abs(k2 - target_result) < TOLERANCE_B:
                return conductivity

            # Adjust conductivity for the next iteration
            STEP = 0.001  # You can adjust the step size as needed
            conductivity -= STEP

    @classmethod
    def from_length(
        cls,
        display_name: str,
        length_m: float,
        pipe_material: PhxHotWaterPipingMaterial,
        pipe_diameter_m: float,
    ) -> "PhxPipeSegment":
        """Create a Pipe Segment from a length value (m)."""
        _l: int = length_m  # type: ignore # untyped LBT

        return cls(
            identifier=str(uuid4()),
            display_name=display_name,
            geometry=LineSegment3D(Point3D(0, 0, 0), Vector3D(_l, 0, 0)),
            pipe_material=pipe_material,
            diameter_m=pipe_diameter_m,
            insulation_thickness_m=0.0,
            insulation_conductivity=0.0,
            insulation_reflective=False,
            insulation_quality=None,
            daily_period=24,
        )


@dataclass
class PhxPipeElement:
    """A Pipe Element / Run made of one or more PhxPipeSegments."""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    identifier: str = field(default_factory=lambda: str(uuid4()))
    display_name: str = "_unnamed_pipe_element_"
    _segments: Dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        PhxPipeElement._count += 1
        self.id_num = PhxPipeElement._count

    @property
    def segments(self) -> List[PhxPipeSegment]:
        return list(self._segments.values())

    @property
    def length_m(self) -> float:
        """Return the total length of all the Pipe Segments."""
        return sum(_.length_m for _ in self.segments)

    @property
    def weighted_pipe_heat_loss_coefficient(self) -> float:
        """Return a length-weighted total heat loss coefficient (W/mk)"""
        weighted_total = 0.0
        for segment in self.segments:
            weighted_total += segment.pipe_heat_loss_coefficient * segment.length_m
        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def weighted_diameter_mm(self) -> float:
        """Return a length-weighted total diameter (mm)"""
        weighted_total = 0.0
        for segment in self.segments:
            weighted_total += segment.diameter_mm * segment.length_m
        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def material(self) -> PhxHotWaterPipingMaterial:
        if not self.segments:
            return PhxHotWaterPipingMaterial.COPPER_K

        materials = list({s.pipe_material for s in self.segments})
        if len(materials) != 1:
            raise ValueError(
                "Error: Pipe Element '{}' has multiple materials: '{}'."
                "Please rebuild the pipe with a single material. {}".format(self.display_name, materials, self.segments)
            )
        else:
            return materials[0]

    @property
    def demand_recirculation(self) -> bool:
        return False

    def add_segment(self, _s: PhxPipeSegment) -> None:
        self._segments[_s.identifier] = _s


@dataclass
class PhxPipeBranch:
    """A Pipe Branch made of one or more Fixture-pipes (PhxPipeElement)."""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    identifier: str = field(default_factory=lambda: str(uuid4()))
    display_name: str = "_unnamed_pipe_branch_"
    pipe_element: PhxPipeElement = field(default_factory=PhxPipeElement)
    fixtures: List[PhxPipeElement] = field(default_factory=list)

    def __post_init__(self) -> None:
        PhxPipeBranch._count += 1
        self.id_num = PhxPipeBranch._count

    def add_fixture(self, _f: PhxPipeElement) -> None:
        self.fixtures.append(_f)


@dataclass
class PhxPipeTrunk:
    """A Pipe Trunk made of one or more Pipe Branches (PhxPipeBranch)."""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    identifier: str = field(default_factory=lambda: str(uuid4()))
    display_name: str = "_unnamed_pipe_trunk_"
    multiplier: int = 1
    pipe_element: PhxPipeElement = field(default_factory=PhxPipeElement)
    branches: List[PhxPipeBranch] = field(default_factory=list)

    def __post_init__(self) -> None:
        PhxPipeTrunk._count += 1
        self.id_num = PhxPipeTrunk._count

    def add_branch(self, _b: PhxPipeBranch) -> None:
        self.branches.append(_b)


# -- Type Alias
AnyPhxPiping = Union[PhxPipeElement, PhxPipeBranch, PhxPipeTrunk]
